/*
 * Copyright (C) 2010 Nokia Corporation and/or its subsidiary(-ies).
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 *           (C) 1999 Antti Koivisto (koivisto@kde.org)
 *           (C) 2001 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2004, 2005, 2006, 2007, 2009, 2010 Apple Inc. All rights reserved.
 *           (C) 2006 Alexey Proskuryakov (ap@nypop.com)
 * Copyright (C) 2010 Google Inc. All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "HTMLSelectElement.h"

#include "AXObjectCache.h"
#include "Attribute.h"
#include "Chrome.h" 
#include "ChromeClient.h" 
#include "Document.h" 
#include "EventHandler.h"
#include "EventNames.h"
#include "ExceptionCodePlaceholder.h"
#include "FormDataList.h"
#include "Frame.h"
#include "HTMLFormElement.h"
#include "HTMLNames.h"
#include "HTMLOptGroupElement.h"
#include "HTMLOptionElement.h"
#include "HTMLOptionsCollection.h"
#include "KeyboardEvent.h"
#include "LocalizedStrings.h"
#include "MouseEvent.h"
#include "NodeRenderingContext.h"
#include "Page.h" 
#include "PlatformConfig.h"
#include "PlatformInput.h"
#include "PlatformMouseEvent.h"
#include "RenderListBox.h"
#include "RenderMenuList.h"
#include "RenderTheme.h"
#include "ScriptEventListener.h"
#include "Settings.h"
#include "SpatialNavigation.h"
#include "MouseEvent.h"
#include "PlatformMouseEvent.h"
#include "PlatformUI.h"

#if PLATFORM(WIIU)
#include "WebViewWiiU.h"
#endif

namespace WebCore {

using namespace HTMLNames;

// Upper limit agreed upon with representatives of Opera and Mozilla.
static const unsigned maxSelectItems = 10000;

HTMLSelectElement::HTMLSelectElement(const QualifiedName& tagName, Document* document, HTMLFormElement* form)
    : HTMLFormControlElementWithState(tagName, document, form)
{
    ASSERT(hasTagName(selectTag));
}

PassRefPtr<HTMLSelectElement> HTMLSelectElement::create(const QualifiedName& tagName, Document* document, HTMLFormElement* form)
{
    ASSERT(tagName.matches(selectTag));
    return adoptRef(new HTMLSelectElement(tagName, document, form));
}

const AtomicString& HTMLSelectElement::formControlType() const
{
    DEFINE_STATIC_LOCAL(const AtomicString, selectMultiple, ("select-multiple"));
    DEFINE_STATIC_LOCAL(const AtomicString, selectOne, ("select-one"));
    return m_data.multiple() ? selectMultiple : selectOne;
}

int HTMLSelectElement::selectedIndex() const
{
    return SelectElement::selectedIndex(m_data, this);
}

void HTMLSelectElement::deselectItems(HTMLOptionElement* excludeElement)
{
    SelectElement::deselectItems(m_data, this, excludeElement);
    setNeedsValidityCheck();
}

void HTMLSelectElement::setSelectedIndex(int optionIndex, bool deselect)
{
    SelectElement::setSelectedIndex(m_data, this, optionIndex, deselect, false, false);
}

void HTMLSelectElement::setSelectedIndexByUser(int optionIndex, bool deselect, bool fireOnChangeNow, bool allowMultipleSelection)
{
    // List box selects can fire onchange events through user interaction, such as
    // mousedown events. This allows that same behavior programmatically.
    if (!m_data.usesMenuList()) {
        updateSelectedState(m_data, this, optionIndex, allowMultipleSelection, false);
        setNeedsValidityCheck();
        if (fireOnChangeNow)
            listBoxOnChange();
        return;
    }

    // Bail out if this index is already the selected one, to avoid running unnecessary JavaScript that can mess up
    // autofill, when there is no actual change (see https://bugs.webkit.org/show_bug.cgi?id=35256 and rdar://7467917 ).
    // Perhaps this logic could be moved into SelectElement, but some callers of SelectElement::setSelectedIndex()
    // seem to expect it to fire its change event even when the index was already selected.
    if (optionIndex == selectedIndex())
        return;
    
    SelectElement::setSelectedIndex(m_data, this, optionIndex, deselect, fireOnChangeNow, true);
}

bool HTMLSelectElement::hasPlaceholderLabelOption() const
{
    // The select element has no placeholder label option if it has an attribute "multiple" specified or a display size of non-1.
    // 
    // The condition "size() > 1" is actually not compliant with the HTML5 spec as of Dec 3, 2010. "size() != 1" is correct.
    // Using "size() > 1" here because size() may be 0 in WebKit.
    // See the discussion at https://bugs.webkit.org/show_bug.cgi?id=43887
    //
    // "0 size()" happens when an attribute "size" is absent or an invalid size attribute is specified.
    // In this case, the display size should be assumed as the default.
    // The default display size is 1 for non-multiple select elements, and 4 for multiple select elements.
    //
    // Finally, if size() == 0 and non-multiple, the display size can be assumed as 1.
    if (multiple() || size() > 1)
        return false;

    int listIndex = optionToListIndex(0);
    ASSERT(listIndex >= 0);
    if (listIndex < 0)
        return false;
    HTMLOptionElement* option = static_cast<HTMLOptionElement*>(listItems()[listIndex]);
    return !listIndex && option->value().isEmpty();
}

bool HTMLSelectElement::valueMissing() const
{
    if (!isRequiredFormControl())
        return false;

    int firstSelectionIndex = selectedIndex();

    // If a non-placeholer label option is selected (firstSelectionIndex > 0), it's not value-missing.
    return firstSelectionIndex < 0 || (!firstSelectionIndex && hasPlaceholderLabelOption());
}

void HTMLSelectElement::listBoxSelectItem(int listIndex, bool allowMultiplySelections, bool shift, bool fireOnChangeNow)
{
    if (!multiple())
        setSelectedIndexByUser(listToOptionIndex(listIndex), true, fireOnChangeNow);
    else {
        updateSelectedState(m_data, this, listIndex, allowMultiplySelections, shift);
        setNeedsValidityCheck();
        if (fireOnChangeNow)
            listBoxOnChange();
    }
}

bool HTMLSelectElement::usesMenuList() const
{
    const Page* page = document()->page();
    RefPtr<RenderTheme> renderTheme = page ? page->theme() : RenderTheme::defaultTheme();
    //if (renderTheme->delegatesMenuListRendering())
        return true;

    return !multiple() && size() <= 1;
}

int HTMLSelectElement::activeSelectionStartListIndex() const
{
    if (m_data.activeSelectionAnchorIndex() >= 0)
        return m_data.activeSelectionAnchorIndex();
    return optionToListIndex(selectedIndex());
}

int HTMLSelectElement::activeSelectionEndListIndex() const
{
    if (m_data.activeSelectionEndIndex() >= 0)
        return m_data.activeSelectionEndIndex();
    return SelectElement::lastSelectedListIndex(m_data, this);
}

unsigned HTMLSelectElement::length() const
{
    return SelectElement::optionCount(m_data, this);
}

void HTMLSelectElement::add(HTMLElement* element, HTMLElement* before, ExceptionCode& ec)
{
    RefPtr<HTMLElement> protectNewChild(element); // make sure the element is ref'd and deref'd so we don't leak it

    if (!element || !(element->hasLocalName(optionTag) || element->hasLocalName(hrTag)))
        return;

    insertBefore(element, before, ec);
    setNeedsValidityCheck();
}

void HTMLSelectElement::remove(int optionIndex)
{
    int listIndex = optionToListIndex(optionIndex);
    if (listIndex < 0)
        return;

    ExceptionCode ec;
    listItems()[listIndex]->remove(ec);
}

void HTMLSelectElement::remove(HTMLOptionElement* option)
{
    if (option->ownerSelectElement() != this)
        return;

    ExceptionCode ec;
    option->remove(ec);
}

String HTMLSelectElement::value() const
{
    const Vector<Element*>& items = listItems();
    for (unsigned i = 0; i < items.size(); i++) {
        if (items[i]->hasLocalName(optionTag) && static_cast<HTMLOptionElement*>(items[i])->selected())
            return static_cast<HTMLOptionElement*>(items[i])->value();
    }
    return "";
}

void HTMLSelectElement::setValue(const String &value)
{
    if (value.isNull())
        return;
    // find the option with value() matching the given parameter
    // and make it the current selection.
    const Vector<Element*>& items = listItems();
    unsigned optionIndex = 0;
    for (unsigned i = 0; i < items.size(); i++) {
        if (items[i]->hasLocalName(optionTag)) {
            if (static_cast<HTMLOptionElement*>(items[i])->value() == value) {
                setSelectedIndex(optionIndex, true);
                return;
            }
            optionIndex++;
        }
    }
}

bool HTMLSelectElement::saveFormControlState(String& value) const
{
    return SelectElement::saveFormControlState(m_data, this, value);
}

void HTMLSelectElement::restoreFormControlState(const String& state)
{
    SelectElement::restoreFormControlState(m_data, this, state);
    setNeedsValidityCheck();
}

void HTMLSelectElement::parseMappedAttribute(Attribute* attr)
{
    if (attr->name() == sizeAttr) {
        int oldSize = m_data.size();
        // Set the attribute value to a number.
        // This is important since the style rules for this attribute can determine the appearance property.
        int size = attr->value().toInt();
        String attrSize = String::number(size);
        if (attrSize != attr->value())
            attr->setValue(attrSize);
        size = max(size, 1);

        // Ensure that we've determined selectedness of the items at least once prior to changing the size.
        if (oldSize != size)
            recalcListItemsIfNeeded();

        m_data.setSize(size);
        setNeedsValidityCheck();
        if (m_data.size() != oldSize && attached()) {
            reattach();
            setRecalcListItems();
        }
    } else if (attr->name() == multipleAttr)
        SelectElement::parseMultipleAttribute(m_data, this, attr);
    else if (attr->name() == accesskeyAttr) {
        // FIXME: ignore for the moment
    } else if (attr->name() == alignAttr) {
        // Don't map 'align' attribute.  This matches what Firefox, Opera and IE do.
        // See http://bugs.webkit.org/show_bug.cgi?id=12072
    } else if (attr->name() == onchangeAttr) {
        setAttributeEventListener(eventNames().changeEvent, createAttributeEventListener(this, attr));
    } else
        HTMLFormControlElementWithState::parseMappedAttribute(attr);
}

bool HTMLSelectElement::isKeyboardFocusable(KeyboardEvent* event) const
{
    // do not focus on select elements if "focus navigation" is off
    if (!LibPlatform::Input::GetInstance()->getInputConfig()->m_bRemoteFocusNav && !LibPlatform::Input::GetInstance()->getInputConfig()->m_bGamePadNav) {
        return false;
    }
    if (renderer())
        return isFocusable();
    return HTMLFormControlElementWithState::isKeyboardFocusable(event);
}

bool HTMLSelectElement::isMouseFocusable() const
{
    if (renderer())
        return isFocusable();
    return HTMLFormControlElementWithState::isMouseFocusable();
}

bool HTMLSelectElement::canSelectAll() const
{
    return !m_data.usesMenuList(); 
}

void HTMLSelectElement::selectAll()
{
    SelectElement::selectAll(m_data, this);
    setNeedsValidityCheck();
}

RenderObject* HTMLSelectElement::createRenderer(RenderArena* arena, RenderStyle*)
{
    if (m_data.usesMenuList())
        return new (arena) RenderMenuList(this);
    return new (arena) RenderListBox(this);
}

bool HTMLSelectElement::appendFormData(FormDataList& list, bool)
{
    return SelectElement::appendFormData(m_data, this, list);
}

int HTMLSelectElement::optionToListIndex(int optionIndex) const
{
    return SelectElement::optionToListIndex(m_data, this, optionIndex);
}

int HTMLSelectElement::listToOptionIndex(int listIndex) const
{
    return SelectElement::listToOptionIndex(m_data, this, listIndex);
}

PassRefPtr<HTMLOptionsCollection> HTMLSelectElement::options()
{
    return HTMLOptionsCollection::create(this);
}

void HTMLSelectElement::recalcListItems(bool updateSelectedStates) const
{
    SelectElement::recalcListItems(const_cast<SelectElementData&>(m_data), this, updateSelectedStates);
}

void HTMLSelectElement::recalcListItemsIfNeeded()
{
    if (m_data.shouldRecalcListItems())
        recalcListItems();
}

void HTMLSelectElement::childrenChanged(bool changedByParser, Node* beforeChange, Node* afterChange, int childCountDelta)
{
    setRecalcListItems();
    setNeedsValidityCheck();
    HTMLFormControlElementWithState::childrenChanged(changedByParser, beforeChange, afterChange, childCountDelta);
    
    if (AXObjectCache::accessibilityEnabled() && renderer())
        renderer()->document()->axObjectCache()->childrenChanged(renderer());
}

void HTMLSelectElement::setRecalcListItems()
{
    SelectElement::setRecalcListItems(m_data, this);

    if (!inDocument())
        m_collectionInfo.reset();
}

void HTMLSelectElement::reset()
{
    SelectElement::reset(m_data, this);
    setNeedsValidityCheck();
}

void HTMLSelectElement::dispatchFocusEvent(PassRefPtr<Node> oldFocusedNode)
{
    SelectElement::dispatchFocusEvent(m_data, this);
    HTMLFormControlElementWithState::dispatchFocusEvent(oldFocusedNode);
}

void HTMLSelectElement::dispatchBlurEvent(PassRefPtr<Node> newFocusedNode)
{
    SelectElement::dispatchBlurEvent(m_data, this);
    HTMLFormControlElementWithState::dispatchBlurEvent(newFocusedNode);
}

void HTMLSelectElement::defaultEventHandler(Event* event)
{
    #ifdef PLATFORM (WIIU)

    static bool global_down = false;
    static bool global_up = false;
    
    if(event->type() == eventNames().mousedownEvent) {
        global_up = false;
        global_down = true;
    }
    
    if(event->type() == eventNames().mouseupEvent)   {
        global_up = true;
    }

    if(event->type() == eventNames().mouseoutEvent || event->type() == eventNames().mouseoverEvent) {
        global_down = false;
        global_up = false;
        return;
    }
    
    if (((event->type() == eventNames().keydownEvent) && static_cast<KeyboardEvent*>(event)->keyIdentifier() == "Enter") 
		|| (global_down && global_up  && event->isMouseEvent() && static_cast<MouseEvent*>(event)->button() == LeftButton)) 
	{
        
        // Don't handle the events if webpage is loading a new page
        if (LibPlatform::ApplicationConfig::GetExtendedDisplay() || document()->page()->chrome()->client()->IsDisplayTV())
        {
            WebView *pView = (WebView*)WebViewTV();
            if (pView && pView->GetEventState() == WEBVIEW_EVENT_UNINIT)
                return;
        }
        else
        {
            WebView *pView = (WebView*)WebViewGamePad(0);
            if (pView && pView->GetEventState() == WEBVIEW_EVENT_UNINIT)
                return;
        }

        global_down = false;
        global_up = false;
        // SelectElement::defaultEventHandler(m_data, this, event, form());
		
        // Error message is displayed on the GamePad and Look At GamePad is displayed on TV
		LibPlatform::Error::ScreenType display = LibPlatform::Error::DISPLAY_GAMEPAD;           
 
        switch (LibPlatform::ApplicationConfig::GetSelectElementEnabledTarget()) {
             case DISPLAY_TYPE_TV:
                if (LibPlatform::UI::MenuListIsOpen(true)) {
                    if (LibPlatform::ApplicationConfig::GetSelectElementMode() == LibPlatform::MODE_TYPE_NON_MODAL) {
                        // if non-modal show error
                        const std::wstring& messageString = LibPlatform::Error::GetStringFromCustomText(CHECK_GAMEPAD_DIALOG_ARRAY_INDEX);                        
                        LibPlatform::Error::GetInstance()->EnqueueError(LibPlatform::Error::USER, messageString.data(), L"OK", 0, 0, display, LibPlatform::Error::eAct_Nothing, 0, false);
                    }  else if (LibPlatform::ApplicationConfig::GetSelectElementMode() == LibPlatform::MODE_TYPE_MODAL) {
                        // display will be disabled
                        return;
                    }
                    return;
                }
                break;
             case DISPLAY_TYPE_GAMEPAD0:
                if (LibPlatform::UI::MenuListIsOpen(true)) {
                    if (LibPlatform::ApplicationConfig::GetSelectElementMode() == LibPlatform::MODE_TYPE_NON_MODAL) {
                        // if non-modal show error
                        const std::wstring& messageString = LibPlatform::Error::GetStringFromCustomText(CHECK_GAMEPAD_DIALOG_ARRAY_INDEX);                        
                        LibPlatform::Error::GetInstance()->EnqueueError(LibPlatform::Error::USER, messageString.data(), L"OK", 0, 0, display, LibPlatform::Error::eAct_Nothing, 0, false );
                    } else if (LibPlatform::ApplicationConfig::GetSelectElementMode() == LibPlatform::MODE_TYPE_MODAL) {
                        // display will be disabled
                        return;
                    }
                    return;
                }
                break;
             case DISPLAY_TYPE_ALL:
                // shared
                if ( LibPlatform::UI::MenuListIsOpen(document()->page()->chrome()->client()->IsDisplayTV()) ) {
                    return;
                }            
                break;
        }


           // std::vector< LibPlatform::UI::MenuListItem > menuItems;
            const Vector<Element*>& items = listItems();
            int availableOptionCount = 0;
            int lastAvailableOptionIndex = -1;
            int numberOfItems = items.size(); 
            bool reportError = false; 
            m_menuItems.clear(); 
            
            if (numberOfItems > LibPlatform::UI::MAX_MENU_ITEMS) {
                numberOfItems = LibPlatform::UI::MAX_MENU_ITEMS;
				reportError = true; 
            }
                    
            for (unsigned i = 0; i < numberOfItems; i++) {
                Element* element = items[i];

                std::wstring label;
                bool isGroup = false;
                bool isSelected = false;
                bool isDisabled = !static_cast<HTMLFormControlElement*>(element)->isEnabledFormControl();

                if ( isHTMLOptGroupElement(element) ) {
                    HTMLOptGroupElement* opt = toHTMLOptGroupElement(element);
                    // OSReport("--- Option group: %s\n", opt->groupLabelText().utf8().data() );

                    if (opt->groupLabelText().length() > 0) 
                        label = wstring((wchar_t *)opt->groupLabelText().characters(), 0, opt->groupLabelText().length());
                      
                    isGroup = true;
                }
                if ( isHTMLOptionElement(element) ) {
                    HTMLOptionElement* opt = toHTMLOptionElement(element);
                    // OSReport("----- Option element: %s\n", opt->text().utf8().data() );

                    isSelected = opt->selected();

                    if (opt->text().length() > 0) {
                    	String optionLabel = opt->textIndentedToRespectGroupLabel();
						label = wstring((wchar_t *)optionLabel.characters(), 0, optionLabel.length());
                    }

                    if ( !isDisabled ) {
                        ++availableOptionCount;
                        lastAvailableOptionIndex = i;
                    }
                }

                
                m_menuItems.push_back( LibPlatform::UI::MenuListItem(label, isGroup, isSelected, isDisabled) );
            }

            
           if (reportError && LibPlatform::ApplicationConfig::GetSelectElementEnabledTarget() != LibPlatform::DISPLAY_TYPE_NONE) {
				const std::wstring& msg = LibPlatform::Error::GetStringFromCustomText(MENU_LIST_TOO_MANY_OPTIONS_ARRAY_INDEX);
                const wchar_t* cmsg = msg.c_str();
				//Open Non-Scrolling Alert
				LibPlatform::UI::SetFocusNavActive(event->type() == eventNames().keydownEvent);
				LibPlatform::UI::SetMenuListAlertOpen(true);
				
                // Manually Force Redraw of webview because Alert With No Scroll Dialog will pause webkit and messages cannot be processed
                // A redraw is necessary because the button animation needs to be updated to the released state. 
                // If not, the button will remain in a pressed animation state when the dialog is open.
                WebView *pView = (WebView*)WebViewTV();
                if (pView)
                    pView->redraw();

                LibPlatform::UI::OpenAlertNoScroll((const wchar_t*)msg.c_str(), msg.length(), display);
				LibPlatform::UI::OpenMenuList(
                    m_menuItems,
                    multiple(),
                    (void*)this,
                    document()->page()->chrome()->client()->IsDisplayTV(),
                    this->isDirectionRTL(),
                    event->type() == eventNames().keydownEvent);
				LibPlatform::UI::SetMenuListAlertOpen(false);
                return; 
            }
                
            
            if ( availableOptionCount > 1 ) {
                LibPlatform::UI::OpenMenuList(
                    m_menuItems,
                    multiple(),
                    (void*)this,
                    document()->page()->chrome()->client()->IsDisplayTV(),
                    this->isDirectionRTL(),
                    event->type() == eventNames().keydownEvent);
            }
            else {
                std::wstring message;
                
                if ( availableOptionCount == 0 ) {
                    message = LibPlatform::Error::GetStringFromCustomText(MENU_LIST_NO_SELECTABLE_OPTIONS_ARRAY_INDEX);
                }
                else {
                    ASSERT(lastAvailableOptionIndex >= 0);
                    LibPlatform::UI::MenuListItem item = m_menuItems[lastAvailableOptionIndex];
                    message = LibPlatform::Error::GetStringFromCustomText(MENU_LIST_YOU_CAN_ONLY_SELECT_ARRAY_INDEX);
                    std::wstring itemStr(L"<ITEM>");
                    size_t foundIdx = message.find(itemStr);
                    if (foundIdx != std::string::npos) {
                        // replace <item> with item.label
                        message.replace(foundIdx, itemStr.length(), item.label);
                    } else {
                        // append item.label
                        message += L"\n\"";
                        message += item.label;
                        message += L".\"";
                    }
                }
			
				LibPlatform::UI::SetFocusNavActive(event->type() == eventNames().keydownEvent);
                
                // Manually Force Redraw of webview because Alert With No Scroll Dialog will pause webkit and messages cannot be processed
                // A redraw is necessary because the button animation needs to be updated to the released state. 
                // If not, the button will remain in a pressed animation state when the dialog is open.
                WebView *pView = (WebView*)WebViewTV();
                if (pView)
                    pView->redraw();

				LibPlatform::UI::OpenAlertNoScroll((const wchar_t*)message.c_str(), message.length(), display);
				
                HTMLSelectElement::setSelectedIndex( listToOptionIndex(lastAvailableOptionIndex) );
            }

    }
    else {
        if (!renderer())
            return;

		if (!isEnabledFormControl()) {
            HTMLFormControlElementWithState::defaultEventHandler(event);  
            return;
        }

        if (event->defaultHandled())
            return;

        if (event->type() == eventNames().keypressEvent && event->isKeyboardEvent()) {
            KeyboardEvent* keyboardEvent = static_cast<KeyboardEvent*>(event);
            if (!keyboardEvent->ctrlKey() && !keyboardEvent->altKey() && !keyboardEvent->metaKey()) {
                event->setDefaultHandled();
                return;
            }
        }
        HTMLFormControlElementWithState::defaultEventHandler(event);
    }
#else     
    if (!renderer())
        return;

    if (!isEnabledFormControl()) {
    HTMLFormControlElementWithState::defaultEventHandler(event);  
        return;
    }

    if (event->defaultHandled())
        return;

    if (event->type() == eventNames().keypressEvent && event->isKeyboardEvent()) {
        KeyboardEvent* keyboardEvent = static_cast<KeyboardEvent*>(event);
        if (!keyboardEvent->ctrlKey() && !keyboardEvent->altKey() && !keyboardEvent->metaKey()) {
            event->setDefaultHandled();
            return;
        }
    }
    HTMLFormControlElementWithState::defaultEventHandler(event);
#endif  
}

#if PLATFORM(WIIU)

void HTMLSelectElement::handleMenuClose( const std::list<int>& indexes ) {
    if ( multiple() ) {
        deselectItems();

        std::list<int>::const_iterator it = indexes.begin();
        for ( ; it != indexes.end(); it++ ) {
            // SelectElement::setSelectedIndex(m_data, this, *it, false, false, multiple());
            // SelectElement::setSelectedIndex(m_data, this, listToOptionIndex(*it), false);
			listBoxSelectItem( *it, true, false, true );		// TODO: verify that this is a good replacement for the above
        }
    }
    else {
        ASSERT ( indexes.size() > 0 );
        // SelectElement::setSelectedIndex(m_data, this, listToOptionIndex(indexes.front()) );
			listBoxSelectItem(indexes.front(), false, false, true );	// TODO: verify that this is a good replacement for the above
    }
    if ( usesMenuList() ) {
//        menuListOnChange();
		//dispatchChangeEventForMenuList();
    }
}

void HTMLSelectElement::FiveHundredErrorCallback_MainThread(void *pUserValue)
{
    HTMLSelectElement * element = static_cast<HTMLSelectElement *>(pUserValue); 
    LibPlatform::UI::OpenMenuList( 
		element->m_menuItems, 
		element->multiple(), 
		pUserValue, 
		element->document()->page()->chrome()->client()->IsDisplayTV(), 
		element->isDirectionRTL(), 
		LibPlatform::Error::GetInstance()->IsSelectCursorActive());
}

LibPlatform::Error::PostErrorAction HTMLSelectElement::FiveHundredErrorCallback(int dialogButton, void *pUserValue)
{
    // DOM access requires being executed on the main thread
    WTF::callOnMainThread(FiveHundredErrorCallback_MainThread, pUserValue);
    return LibPlatform::Error::eAct_Nothing; 
}

#endif

void HTMLSelectElement::setActiveSelectionAnchorIndex(int index)
{
    SelectElement::setActiveSelectionAnchorIndex(m_data, this, index);
}

void HTMLSelectElement::setActiveSelectionEndIndex(int index)
{
    SelectElement::setActiveSelectionEndIndex(m_data, index);
}

void HTMLSelectElement::updateListBoxSelection(bool deselectOtherOptions)
{
    SelectElement::updateListBoxSelection(m_data, this, deselectOtherOptions);
    setNeedsValidityCheck();
}

void HTMLSelectElement::menuListOnChange()
{
    SelectElement::menuListOnChange(m_data, this);
}

void HTMLSelectElement::listBoxOnChange()
{
    SelectElement::listBoxOnChange(m_data, this);
}

void HTMLSelectElement::saveLastSelection()
{
    SelectElement::saveLastSelection(m_data, this);
}

void HTMLSelectElement::accessKeyAction(bool sendToAnyElement)
{
    focus();
    dispatchSimulatedClick(0, sendToAnyElement);
}

void HTMLSelectElement::accessKeySetSelectedIndex(int index)
{
    SelectElement::accessKeySetSelectedIndex(m_data, this, index);
}
    
void HTMLSelectElement::setMultiple(bool multiple)
{
    bool oldMultiple = this->multiple();
    int oldSelectedIndex = selectedIndex();
    setAttribute(multipleAttr, multiple ? "" : 0);

    // Restore selectedIndex after changing the multiple flag to preserve
    // selection as single-line and multi-line has different defaults.
    if (oldMultiple != this->multiple())
        setSelectedIndex(oldSelectedIndex);
}

void HTMLSelectElement::setSize(int size)
{
    setAttribute(sizeAttr, String::number(size));
}

Node* HTMLSelectElement::namedItem(const AtomicString& name)
{
    return options()->namedItem(name);
}

Node* HTMLSelectElement::item(unsigned index)
{
    return options()->item(index);
}

void HTMLSelectElement::setOption(unsigned index, HTMLOptionElement* option, ExceptionCode& ec)
{
    ec = 0;
    if (index > maxSelectItems - 1)
        index = maxSelectItems - 1;
    int diff = index  - length();
    HTMLElement* before = 0;
    // out of array bounds ? first insert empty dummies
    if (diff > 0) {
        setLength(index, ec);
        // replace an existing entry ?
    } else if (diff < 0) {
        before = toHTMLElement(options()->item(index+1));
        remove(index);
    }
    // finally add the new element
    if (!ec) {
        add(option, before, ec);
        if (diff >= 0 && option->selected())
            setSelectedIndex(index, !m_data.multiple());
    }
}

void HTMLSelectElement::setLength(unsigned newLen, ExceptionCode& ec)
{
    ec = 0;
    if (newLen > maxSelectItems)
        newLen = maxSelectItems;
    int diff = length() - newLen;

    if (diff < 0) { // add dummy elements
        do {
            RefPtr<Element> option = document()->createElement(optionTag, false);
            ASSERT(option);
            add(toHTMLElement(option.get()), 0, ec);
            if (ec)
                break;
        } while (++diff);
    } else {
        const Vector<Element*>& items = listItems();

        // Removing children fires mutation events, which might mutate the DOM further, so we first copy out a list
        // of elements that we intend to remove then attempt to remove them one at a time.
        Vector<RefPtr<Element> > itemsToRemove;
        size_t optionIndex = 0;
        for (size_t i = 0; i < items.size(); ++i) {
            Element* item = items[i];
            if (item->hasLocalName(optionTag) && optionIndex++ >= newLen) {
                ASSERT(item->parentNode());
                itemsToRemove.append(item);
            }
        }

        for (size_t i = 0; i < itemsToRemove.size(); ++i) {
            Element* item = itemsToRemove[i].get();
            if (item->parentNode()) {
                item->parentNode()->removeChild(item, ec);
            }
        }
    }
    setNeedsValidityCheck();
}

void HTMLSelectElement::scrollToSelection()
{
    SelectElement::scrollToSelection(m_data, this);
}

void HTMLSelectElement::insertedIntoTree(bool deep)
{
    SelectElement::insertedIntoTree(m_data, this);
    HTMLFormControlElementWithState::insertedIntoTree(deep);
}

bool HTMLSelectElement::isRequiredFormControl() const
{
    return required();
}

#ifdef PLATFORM (WIIU)
static const Element* parentHTMLElement(const Element* element)
{
    while (element) {
        element = element->parentElement();
        if (element && element->isHTMLElement())
            return element;
    }
    return 0;
}

String HTMLSelectElement::directionForOptionData() const
{
    for (const Element* element = this; element; element = parentHTMLElement(element)) {
        const AtomicString& dirAttributeValue = element->fastGetAttribute(dirAttr);

        if (dirAttributeValue.isNull())
            continue;

        //if (equalIgnoringCase(dirAttributeValue, "rtl") || equalIgnoringCase(dirAttributeValue, "ltr")) {
        //    return dirAttributeValue;
        //}

        if (equalIgnoringCase(dirAttributeValue, "rtl")) {
            return dirAttributeValue;
        }

        if (equalIgnoringCase(dirAttributeValue, "ltr")) {
            return dirAttributeValue;
        }

        if (equalIgnoringCase(dirAttributeValue, "auto")) {
            bool isAuto;
            TextDirection textDirection = static_cast<const HTMLElement*>(element)->directionalityIfhasDirAutoAttribute(isAuto);
            return textDirection == RTL ? "rtl" : "ltr";
        }
    }

    return "ltr";
}

bool HTMLSelectElement::isDirectionRTL() const
{
        String dir = directionForOptionData();
        return (dir == "rtl");
}
#endif

} // namespace
