/*
 * Copyright (C) 2006 Apple Computer, Inc.
 * Copyright (C) 2006 Michael Emmel mike.emmel@gmail.com
 * Copyright (C) 2007, 2008 Alp Toker <alp@atoker.com>
 * Copyright (C) 2007 Holger Hans Peter Freyther
 * Copyright (C) 2009, 2010 Igalia S.L.
 * Copyright (c) 2012 NINTENDO.
 * All rights reserved.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 */

#include "config.h"
#include "FontPlatformData.h"

#include "NotImplemented.h"
#include <algorithm>
#include <stdio.h>
#include "GraphicsTypes.h"
#include "GraphicsContextPlatformPrivate.h"
#include <wkgfx_Misc.h>

#include "FontDescription.h"
#include "wtf/wiiu/StatisticsWiiU.h"
#include <wtf/text/WTFString.h>

namespace WebCore {

FontPlatformData::FontPlatformData(struct font_config_handle *handle, font_param *pattern, const FontDescription& fontDescription)
    : m_size(fontDescription.computedPixelSize())
    , m_syntheticBold(false)
    , m_syntheticOblique(false)
    , m_fixedWidth(false)
    , m_scaledFont(0)
{
	m_handle = handle;
	if(pattern->family_name)
	{
		m_pattern.family_name = static_cast<char*>(malloc(sizeof(char)*(strlen(pattern->family_name)+1)));
		strcpy(m_pattern.family_name, pattern->family_name);
	}
	m_pattern.style_flags = pattern->style_flags;
	m_pattern.characters = pattern->characters;
	m_pattern.num_chars = pattern->num_chars;
	m_pattern.font_weight = pattern->font_weight;

    m_scaledFont = font_face_t::create();
    if(m_scaledFont)
        m_scaledFont->use();
}

    FontPlatformData::FontPlatformData(float size, bool bold, bool italic)
    : m_size(size)
    , m_syntheticBold(bold)
    , m_syntheticOblique(italic)
    , m_fixedWidth(false)
    , m_scaledFont(0)
{
    m_scaledFont = NULL;
    m_pattern.family_name = NULL;
    m_pattern.style_flags = 0;
    m_pattern.characters = NULL;
    m_pattern.num_chars = 0;
    m_pattern.font_weight = 0;
}

FontPlatformData::FontPlatformData(font_face_t *fontFace, char * familyName,  float size, bool bold, bool italic)
    : m_size(size)
    , m_syntheticBold(bold)
    , m_syntheticOblique(italic)
    , m_fixedWidth(false)
    , m_scaledFont(fontFace)
{
    if (bold)
        m_pattern.font_weight = 100;
    else
        m_pattern.font_weight = 0;

    if (italic)
        m_pattern.style_flags = 1;
    else
        m_pattern.style_flags = 0;

	if(familyName)
	{
		m_pattern.family_name = static_cast<char*>(malloc(sizeof(char)*(strlen(familyName)+1)));
		strcpy(m_pattern.family_name, familyName);
	}

	if (m_scaledFont)
		m_scaledFont->use();
	m_handle = NULL;
}

FontPlatformData& FontPlatformData::operator=(const FontPlatformData& other)
{
    // Check for self-assignment.
    if (this == &other)
        return *this;

    m_size = other.m_size;
    m_syntheticBold = other.m_syntheticBold;
    m_syntheticOblique = other.m_syntheticOblique;
    m_fixedWidth = other.m_fixedWidth;
    m_pattern = other.m_pattern;

    //CRANK: DLM should possibly wrap this in a smart pointer?
	//TF: almost did that; but: no TR1 or boost!
	m_handle = other.m_handle;

	if(other.m_pattern.family_name)
	{
		m_pattern.family_name = static_cast<char*>(malloc(sizeof(char)*(strlen(other.m_pattern.family_name)+1)));
		strcpy(m_pattern.family_name, other.m_pattern.family_name);
	}
	else
	{
		m_pattern.family_name = 0;
	}

    m_pattern.style_flags = other.m_pattern.style_flags;
    m_pattern.characters = other.m_pattern.characters;
    m_pattern.num_chars = other.m_pattern.num_chars;
    m_pattern.font_weight = other.m_pattern.font_weight;

	if (m_scaledFont)
		m_scaledFont->destroy();
    m_scaledFont = other.m_scaledFont;
	if (m_scaledFont)
		m_scaledFont->use();

    return *this;
}

FontPlatformData::FontPlatformData(const FontPlatformData& other)
    : m_scaledFont(0)
{
    *this = other;
}

FontPlatformData::FontPlatformData(const FontPlatformData& other, float size)
    : m_scaledFont(0)
{
    m_size = size;
    *this = other;
}

FontPlatformData::~FontPlatformData()
{
	if (m_scaledFont) {
		m_scaledFont->destroy();
		m_scaledFont = NULL;
		WTF::STATS_WIIU_DEC_FONTCACHE_COUNT();
	}

	if(m_pattern.family_name)
	{
		free(m_pattern.family_name);
		m_pattern.family_name = 0;
	}
}

bool FontPlatformData::isFixedPitch()
{
    return m_fixedWidth;
}

bool FontPlatformData::operator==(const FontPlatformData& other) const
{

    if (isHashTableDeletedValue() != other.isHashTableDeletedValue()) {
        return false;
    }

	bool pattern_result = false;
	if (m_pattern.family_name == NULL && other.m_pattern.family_name == NULL) {
		pattern_result = true;
	} else if (m_pattern.family_name == hashTableDeletedFontValue() && other.m_pattern.family_name == hashTableDeletedFontValue()) {
		pattern_result = true;
	} else if (m_pattern.family_name && m_pattern.family_name != hashTableDeletedFontValue() && other.m_pattern.family_name && other.m_pattern.family_name != hashTableDeletedFontValue()) {
		pattern_result = (strcmp(m_pattern.family_name, other.m_pattern.family_name) == 0)
            && m_pattern.font_weight == other.m_pattern.font_weight
            && m_pattern.style_flags == other.m_pattern.style_flags;
	}

    if (!pattern_result) {
        return false;
    }

    bool result = m_size == other.m_size
    	     && m_syntheticOblique == other.m_syntheticOblique
    	     && m_syntheticBold == other.m_syntheticBold;

    return result;
}

#ifndef NDEBUG
String FontPlatformData::description() const
{
    return String();
}
#endif

}
