/*
 * Copyright (C) 1999 Lars Knoll (knoll@kde.org)
 * Copyright (C) 1999 Antti Koivisto (koivisto@kde.org)
 * Copyright (C) 2000 Dirk Mueller (mueller@kde.org)
 * Copyright (C) 2003, 2006 Apple Inc. All rights reserved.
 * Copyright (C) 2008 INdT - Instituto Nokia de Tecnologia
 * Copyright (C) 2009-2010 ProFUSION embedded systems
 * Copyright (C) 2009-2010 Samsung Electronics
 * Copyright (c) 2012 NINTENDO.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Library General Public
 * License as published by the Free Software Foundation; either
 * version 2 of the License, or (at your option) any later version.
 *
 * This library is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Library General Public License for more details.
 *
 * You should have received a copy of the GNU Library General Public License
 * along with this library; see the file COPYING.LIB.  If not, write to
 * the Free Software Foundation, Inc., 51 Franklin Street, Fifth Floor,
 * Boston, MA 02110-1301, USA.
 *
 */

#include "config.h"
#include "Font.h"

#include "GraphicsContext.h"
#include "GraphicsContextPlatformPrivate.h"
#include "NotImplemented.h"
#include "SimpleFontData.h"

namespace WebCore {

void Font::drawComplexText(GraphicsContext *context, const TextRun& run, const FloatPoint& point, int from, int to) const
{
	drawSimpleText(context, run, point, from, to);
}

void Font::drawEmphasisMarksForComplexText(GraphicsContext* /* context */, const TextRun& /* run */, const AtomicString& /* mark */, const FloatPoint& /* point */, int /* from */, int /* to */) const
{
	notImplemented();
}

bool Font::canReturnFallbackFontsForComplexText()
{
	return false;
}

bool Font::canExpandAroundIdeographsInComplexText()
{
	return false;
}

float Font::floatWidthForComplexText(const TextRun& run, HashSet<const SimpleFontData*>* fallbackFonts, GlyphOverflow *overflow ) const
{
	return floatWidthForSimpleText(run, 0, fallbackFonts, overflow);
}

int Font::offsetForPositionForComplexText(const TextRun& run, float xFloat, bool includePartialGlyphs) const
{
	return offsetForPositionForSimpleText(run, xFloat, includePartialGlyphs);
}

FloatRect Font::selectionRectForComplexText(const TextRun& run, const FloatPoint& point, int h, int from, int to) const
{
	return selectionRectForSimpleText(run, point, h, from, to);
}

// --------------------------------------------------------------------------------------------------------
/**
    Stroke glyphs.	
    We use freetype to accumulate the glyph vectors into a path, and then stroke the path. We don't do any caching.  
    @param context The graphics context.
    @param font The font to use.
    @param glyphBuffer A buffer containing glyphs to render.
    @param from The starting point in the glyphBuffer at which to begin rendering.
    @param numGlyphs The number of glyphs in the glyphBuffer to render.
    @param point The point to start writing text.
*/
void strokeGlyphs(GraphicsContext* context, const SimpleFontData* font, const GlyphBuffer& glyphBuffer,
					  int from, int numGlyphs, const FloatPoint& point)
	{
	GraphicsContextPlatformPrivate *priv = context->platformContext();

    if (!context->makeCurrent())
        return;

#if PLATFORM(WIIU)
    priv->strokeGlyph(font, glyphBuffer, from, numGlyphs, point,context->strokeThickness(), context->strokeColor());
#else

    context->save();
    context->concatCTM(AffineTransform(1.0, 0, 0, -1.0, point.x(), point.y()));

	for (signed i = 0; i < numGlyphs; ++i)
	{
       Path				p;

       if(!priv->setStrokeGlyph(font))
			continue;

        context->strokePath(p);
        context->translate(FloatSize(glyphBuffer.advances(from + i)->width(), 0));
    }
    context->restore();
#endif
}

// --------------------------------------------------------------------------------------------------------
/**
	Draw a horizontal row of glyphs in a given context using a given font
*/
void Font::drawGlyphs(GraphicsContext* context, const SimpleFontData* font, const GlyphBuffer& glyphBuffer,
					  int from, int numGlyphs, const FloatPoint& point) const
			{
	GraphicsContextPlatformPrivate *priv = context->platformContext();
	if (!context->makeCurrent())
		return;
			
    // Any shadows?
    if (priv->isShadowed())
			{
        // Yes. Switch them of so we can render them without, well, shadows...
        priv->disableShadow(); 
				
        // Get handler for blured shadows
        ShadowBlur &shadow = priv->getShadowBlurHandler();

        // Any blured shadows in order?
        if (!shadow.mustUseShadowBlur(context))
				{
            // Use the simple font shadow: just rerender the glyphs in the shadow color & with an offset
				
            context->save();
            context->setFillColor(priv->shadowColor(), ColorSpace());
            context->translate(priv->shadowWidth(), priv->shadowHeight());
            drawGlyphs(context, font, glyphBuffer, from, numGlyphs, point);
            context->restore();
			}
        else
        {
            // Generate properly blured font shadows

            float width = 0;
            for (int i = 0; i < numGlyphs; ++i) width += glyphBuffer.advances(from + i)->width();
            float height = font->fontMetrics().floatHeight() + 1;
			
            FloatRect extents(point.x(), point.y() - font->fontMetrics().floatAscent(), width, height);

            //The case where beginShadowLayer returns 0 is the case where the extents are completely
            //clipped out, so there is no point in trying a fallback. Like a lot of WebKit, if a 
            //memory allocation failed, we would have crashed by now anyway...
            if (GraphicsContext *shadowContext = shadow.beginShadowLayer(context, extents))
			{
				shadowContext->setFillColor(priv->shadowColor(), ColorSpace());
				shadowContext->setTextDrawingMode(TextModeFill);
				//see if we should also stroke the glyphs
				if (context->textDrawingMode() & TextModeStroke)
					strokeGlyphs(context, font, glyphBuffer, from, numGlyphs, point);

                shadowContext->platformContext()->drawGlyphs(font, glyphBuffer, from, numGlyphs, point, priv->shadowColor());


			
                shadow.endShadowLayer(context);
            }
				}
			
        priv->enableShadow(); 
    } 
	
				
    if (context->textDrawingMode() == TextModeInvisible)
        return;

   
	priv->drawGlyphs(font, glyphBuffer, from, numGlyphs, point, context->fillColor());
    //see if we should also stroke the glyphs
    if (context->textDrawingMode() & TextModeStroke)
        strokeGlyphs(context, font, glyphBuffer, from, numGlyphs, point);

   }
	
} // namespace WebCore

