/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without 
 * modification, are permitted provided that the following conditions 
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright 
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the 
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY 
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED 
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE 
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY 
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES 
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; 
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED 
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT 
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF 
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */
//////////////////////////////////////////////////////////////////////////
//	Name	: MediaPlayerPrivateUnix.h
//
//	Purpose	: This is the unix WiiU version of the MediaPlayerPrivate
//				interface
//////////////////////////////////////////////////////////////////////////
#pragma once

#include "Timer.h"
#include "MediaPlayerPrivate.h"
#include "renderer/WiiURenderer.h"
#include "renderer/MediaFoundation/MFVSGCCShim.h"
#include "VideoLayerWiiU.h" 

namespace WebCore
{

class MediaPlayerPrivateUnix;
class MediaPlayerPrivateUnixTimerObject
{
	public:
		MediaPlayerPrivateUnixTimerObject() : m_Timer(this, &MediaPlayerPrivateUnixTimerObject::UpdateVideo), m_pOwner(0){}
		void UpdateVideo(WebCore::Timer<MediaPlayerPrivateUnixTimerObject>* vTimer);

	public:
		MediaPlayerPrivateUnix * m_pOwner;
		WebCore::Timer<MediaPlayerPrivateUnixTimerObject> m_Timer;
};

class MediaPlayerPrivateUnix : public MediaPlayerPrivateInterface
{
	public:
		//WebKit registration support.
		static void getSupportedTypes(HashSet<String>& vTypes);
		static void registerMediaEngine(MediaEngineRegistrar);
		static MediaPlayer::SupportsType supportsType(const String& vType, const String& vCodecs);
		static PassOwnPtr<MediaPlayerPrivateInterface> create(MediaPlayer*);

		MediaPlayerPrivateUnix(MediaPlayer * vOwner);
		virtual ~MediaPlayerPrivateUnix();

		//////////////////////////////////////////////////////////////////////////
		///	Name	: load
		///
		///	Purpose	: Opens the given URL and beings downloading the data for
		///				playback.
		//////////////////////////////////////////////////////////////////////////
		virtual void load(const String& vURL);

		//////////////////////////////////////////////////////////////////////////
		///	Name	: cancelLoad
		///
		///	Purpose	: Cancels the action caused by 'load'
		//////////////////////////////////////////////////////////////////////////
		virtual void cancelLoad();

		//////////////////////////////////////////////////////////////////////////
		///	Name	: play
		///
		///	Purpose	: Begins playback of the loaded media.
		//////////////////////////////////////////////////////////////////////////
		virtual void play();

		//////////////////////////////////////////////////////////////////////////
		///	Name	: pause
		///
		///	Purpose	: Pauses playback of the loaded media.
		//////////////////////////////////////////////////////////////////////////
		virtual void pause();

		//////////////////////////////////////////////////////////////////////////
		///	Name	: seek
		///
		///	Purpose	: Seeks to the given time. Has no effect unless 'load' has
		///				first been called.
		//////////////////////////////////////////////////////////////////////////
		virtual void seek(float vTime);

		//////////////////////////////////////////////////////////////////////////
		//	Name	: State Accessors
		//////////////////////////////////////////////////////////////////////////
		virtual bool paused() const {return m_pMediaPlayerShim->IsPaused();}
		virtual bool seeking() const {return m_pMediaPlayerShim->IsSeeking();}
		virtual bool hasAudio() const {return m_pMediaPlayerShim->HasAudio();}
		virtual bool hasVideo() const {return m_pMediaPlayerShim->HasVideo();}
		virtual float duration() const {return m_pMediaPlayerShim->GetDuration();}
		virtual float currentTime() const {return m_pMediaPlayerShim->GetCurrentTime();}
		virtual IntSize naturalSize() const;
		virtual unsigned bytesLoaded() const {return m_pMediaPlayerShim->GetNumBytesLoaded();}
		virtual float maxTimeSeekable() const {return m_pMediaPlayerShim->GetMaxSeekableTime();}

		//////////////////////////////////////////////////////////////////////////
		//	Name	: Modifiers
		//////////////////////////////////////////////////////////////////////////
		virtual void setSize(const IntSize& vSize) {m_OutputSize.Set(vSize.width(), vSize.height());}
		virtual void setVolume(float vVolume);
		virtual void setVisible(bool vIsVisible) {m_IsVisible = vIsVisible;}

		//We'll just always say our network and ready state are fine.
		virtual MediaPlayer::NetworkState networkState() const {return MediaPlayer::Loaded;}
		virtual MediaPlayer::ReadyState readyState() const {return MediaPlayer::HaveEnoughData;}

		virtual PassRefPtr<TimeRanges> buffered() const;
		virtual void paint(GraphicsContext * vContext, const IntRect& vDest);

		virtual void paintCurrentFrameInContext(GraphicsContext* c, const IntRect& r) { paint(c, r); }
		virtual void setPreload(MediaPlayer::Preload) { };
		virtual bool hasAvailableVideoFrame() const { return readyState() >= MediaPlayer::HaveCurrentData; }


#if ENABLE(PLUGIN_PROXY_FOR_VIDEO)
		//Don't need to implement any of these
		virtual void deliverNotification(MediaPlayerProxyNotificationType) = 0;
		virtual void setMediaPlayerProxy(WebMediaPlayerProxy*) = 0;
		virtual void setControls(bool) { }
		virtual void enterFullscreen() { }
		virtual void exitFullscreen() { }
#endif

#if USE(ACCELERATED_COMPOSITING)
		// whether accelerated rendering is supported by the media engine for the current media.
		virtual bool supportsAcceleratedRendering() const;
		// called when the rendering system flips the into or out of accelerated rendering mode.
		virtual void acceleratedRenderingStateChanged();
        
        void createVideoLayer();
        bool isReadyForVideoSetup() const;
        void setUpVideoRendering();
        void tearDownVideoRendering();

        enum MediaRenderingMode { MediaRenderingNone, MediaRenderingToContext, MediaRenderingToLayer };
        MediaRenderingMode currentRenderingMode() const;
        MediaRenderingMode preferredRenderingMode() const;
#endif

#pragma region IGNORE ME FOR NOW
		//Don't need to implement any of this stuff
		//////////////////////////////////////////////////////////////////////////
		virtual void setRate(float){};
		virtual void prepareToPlay() { }
		virtual bool hasSingleSecurityOrigin() const { return false; }
		virtual MediaPlayer::MovieLoadType movieLoadType() const { return MediaPlayer::Unknown; }
		virtual void prepareForRendering() { }
		// Time value in the movie's time scale. It is only necessary to override this if the media
		// engine uses rational numbers to represent media time.
		virtual float mediaTimeForTimeValue(float timeValue) const { return timeValue; }
		// Override this if it is safe for HTMLMediaElement to cache movie time and report
		// 'currentTime' as [cached time + elapsed wall time]. Returns the maximum wall time
		// it is OK to calculate movie time before refreshing the cached time.
		virtual double maximumDurationToCacheMediaTime() const { return 0; }
		virtual unsigned decodedFrameCount() const { return 0; }
		virtual unsigned droppedFrameCount() const { return 0; }
		virtual unsigned audioDecodedByteCount() const { return 0; }
		virtual unsigned videoDecodedByteCount() const { return 0; }
		void getSitesInMediaCache(Vector<String>&) { }
		void clearMediaCache() { }
		void clearMediaCacheForSite(const String&) { }
		virtual void setPrivateBrowsingMode(bool) { }

		#if ENABLE(WEB_AUDIO)
		virtual AudioSourceProvider* audioSourceProvider() { return 0; }
		#endif

		#if ENABLE(MEDIA_SOURCE)
		virtual bool sourceAppend(const unsigned char*, unsigned) { return false; }
		virtual void sourceEndOfStream(MediaPlayer::EndOfStreamStatus) { };
		#endif

		virtual PlatformMedia platformMedia() const { return NoPlatformMedia; }

		#if USE(ACCELERATED_COMPOSITING)
		virtual PlatformLayer* platformLayer() const { return m_videoLayer.get(); }
		#endif

		virtual bool supportsFullscreen() const { return false; }
		virtual bool supportsSave() const { return false; }
		virtual float startTime() const { return 0; }
		virtual double initialTime() const { return 0; }
		virtual bool supportsMuting() const { return false; }
		virtual void setMuted(bool) { }
		virtual bool hasClosedCaptions() const { return false; }
		virtual void setClosedCaptionsVisible(bool) { }
		virtual void setPreservesPitch(bool) { }
		virtual bool canLoadPoster() const { return false; }
		virtual void setPoster(const String&) { }
		//////////////////////////////////////////////////////////////////////////
#pragma endregion

	private:
		//////////////////////////////////////////////////////////////////////////
		//	Name	: Copy Constructor
		//////////////////////////////////////////////////////////////////////////
		MediaPlayerPrivateUnix(const MediaPlayerPrivateUnix& vObject) {*this = vObject;}

		//////////////////////////////////////////////////////////////////////////
		//	Name	: Operator=
		//////////////////////////////////////////////////////////////////////////
		MediaPlayerPrivateUnix& operator=(const MediaPlayerPrivateUnix& vObject){return *this;}

		//////////////////////////////////////////////////////////////////////////
		///	Name	: LoadMediaDLL
		///
		///	Purpose	: Loads and initializes the multimedia DLL needed to handle
		///				all of the functionality this class is expected to have.
		//////////////////////////////////////////////////////////////////////////
		MPERROR LoadMediaDLL(const char * vDLLName);

		//////////////////////////////////////////////////////////////////////////
		///	Name	: UnloadMediaDLL
		///
		///	Purpose	: Unloads the currently loaded media DLL.
		//////////////////////////////////////////////////////////////////////////
		void UnloadMediaDLL();

	public:
		bool           m_HasPlaybackStateChanged;
		bool           m_IsVisible;
		void *         m_pMediaDLL;        ///< Handle to the DLL containing the media player this class wraps
		tPoint2D       m_OutputSize;       ///< The size at which video will be rendered, regardless of the source size
		unsigned int   m_VideoTextureID;
		MediaPlayer *  m_pOwner;           ///< The WebKit media player that owns this media player
		tMFVSGCCShim * m_pMediaPlayerShim; ///< The media player that will handle all of the necessary playback operations

		MediaPlayerPrivateUnixTimerObject m_UpdateTimer;
#ifdef USE(ACCELERATED_COMPOSITING)
    RefPtr<VideoLayerWiiU>              m_videoLayer;
#endif 
};

}	//End WebCore



//////////////////////////////////////////////////////////////////////////
///	Name	: LogMediaMessage
///
///	Purpose	: Logs the given message to the media message file.
//////////////////////////////////////////////////////////////////////////
void LogMediaMessage(const char * vMessage, ...);

extern bool g_EnableMediaMessageLogging;
#define LOG_MEDIA_MSG(vMessage, ...) if (g_EnableMediaMessageLogging) LogMediaMessage(vMessage, __VA_ARGS__)
#define LOG_SIMPLE_MEDIA_MSG(vMessage) if (g_EnableMediaMessageLogging) LogMediaMessage(vMessage)
