/*
 * Copyright (C) 2010 Google Inc. All rights reserved.
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions are
 * met:
 *
 *     * Redistributions of source code must retain the above copyright
 * notice, this list of conditions and the following disclaimer.
 *     * Redistributions in binary form must reproduce the above
 * copyright notice, this list of conditions and the following disclaimer
 * in the documentation and/or other materials provided with the
 * distribution.
 *     * Neither the name of Google Inc. nor the names of its
 * contributors may be used to endorse or promote products derived from
 * this software without specific prior written permission.
 *
 * THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
 * "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
 * LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
 * A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
 * OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
 * SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
 * LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
 * DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
 * THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
 * OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "config.h"
#include "PlatformConfig.h"

#if USE(ACCELERATED_COMPOSITING)

#include "VideoLayerWiiU.h"
#include "ImageBuffer.h" 
#include "GraphicsContextPlatformPrivate.h"


const int TILING_SIZE = 2048;

namespace WebCore {

PassRefPtr<VideoLayerWiiU> VideoLayerWiiU::create(GraphicsLayerWiiU* owner)
{
    return adoptRef(new VideoLayerWiiU(owner));
}

VideoLayerWiiU::VideoLayerWiiU(GraphicsLayerWiiU* owner)
    : LayerWiiU(owner), m_player()
{
    pthread_mutexattr_t attr;
    pthread_mutexattr_init(&attr);
    pthread_mutexattr_settype(&attr, PTHREAD_MUTEX_RECURSIVE);
    pthread_mutex_init(&m_player_mutex, &attr);
    pthread_mutexattr_destroy(&attr);
}

VideoLayerWiiU::~VideoLayerWiiU()
{
    pthread_mutex_destroy(&m_player_mutex);
}

void VideoLayerWiiU::cleanupResources()
{
}

void VideoLayerWiiU::setPlayer(MediaPlayer* player)
{
    pthread_mutex_lock(&m_player_mutex);
    // setContents() currently gets called whenever there is any
    // style change that affects the layer even if that change doesn't
    // affect the actual contents of the image (e.g. a CSS animation).
    // With this check in place we avoid unecessary texture uploads.
    if ((m_player == player))
    {
        pthread_mutex_unlock(&m_player_mutex);
        return;
    }

    m_player = player;
    pthread_mutex_unlock(&m_player_mutex);
    m_dirtyRect = IntRect(IntPoint(0, 0), bounds());
    setNeedsDisplay();
}

void VideoLayerWiiU::paintContentLayerIfDirty()
{
    int width  = contentBounds().width();
    int height = contentBounds().height();
    //OS::Report("Paint dirty Layer %s Pos (%fx%f) Size (%dx%d)\n", name().utf8().data(), m_owner->position().x(), m_owner->position().y(), width, height);

    pthread_mutex_lock(&m_player_mutex);
    if (drawsContent()  && m_contentsDirty && (width > 0) && (width < TILING_SIZE) && (height > 0) && (height < TILING_SIZE)) {
        int count = m_owner->incrementRepaintCount();
        //see if we can re-use our current buffer
        if (m_buffer) {
            if (m_buffer->width() != width || m_buffer->height() != height) {
                m_buffer.release();
                m_buffer = ImageBuffer::create((IntSize(width, height)));
            }
        } else {
            m_buffer = ImageBuffer::create((IntSize(width, height)));
        }

        //see if we got an image buffer, it is possible that the allocation could fail
        //if that happens, we'll drop through to the "slow" render path in the paint
        //call.
        if (m_buffer) {
            m_buffer->context()->save();
            m_buffer->context()->clearRect(FloatRect(0.0, 0.0, (float)width, (float)height));
            m_player->paint(m_buffer->context(), IntRect(0, 0, width, height));
            m_buffer->context()->restore();
        }

        m_contentsDirty = false;
        m_dirtyRect = FloatRect();

        if (m_owner->showDebugBorders()) {
            drawDebugBorder(m_buffer->context(), width, height);
        }
    } 
    pthread_mutex_unlock(&m_player_mutex);
}

void VideoLayerWiiU::paintContents(GraphicsContext *ctx)
{
    ASSERT(ctx);
    
    const int renderLayerPadding = 5;

    int contentWidth = bounds().width();
	int contentHeight = bounds().height();
	int image_width  = contentBounds().width();
	int image_height = contentBounds().height();
	int x = 0;
	int y = 0;
    
    int tvResolutionX = 0, tvResolutionY = 0;
    LibPlatform::ApplicationConfig::GetTVDisplayDimensions(tvResolutionX, tvResolutionY);
    
    int widthDifference = contentWidth - image_width;

    if (tvResolutionX >= contentWidth)
    {
        // The content window width is smaller than the TV resolution width
        if (image_width < contentWidth && widthDifference > renderLayerPadding)
	    {
		    x += widthDifference / 2;
	    }
    }
    else
    {
        // The content window width is larger than the TV Resolution width
        // Cap the width to the TV Resolution width
        if (image_width < tvResolutionX)
        {
            x += (tvResolutionX - image_width) / 2;
        }
    }

    if (m_player && m_player->isFullscreen())
    {
        // The content window is smaller than the TV resolution 
        if (tvResolutionY >= contentHeight)
        {
            // Center the video in the content window
            if (image_height < contentHeight)
	        {
		        y += (contentHeight - image_height) / 2;
	        }
        }
        else
        {
            // The content window is larger than the TV Resolution

            // Center the video in screen
            if (image_height < tvResolutionY)
	        {
		        y += (tvResolutionY - image_height) / 2;
	        }
        }
    }
    else
    {
        if (image_height < contentHeight)
	    {
		    y += (contentHeight - image_height) / 2;
	    }
    }

    // the size of the video container can change dynamically seperately from the actual video, 
    //   so we need to create an offset size for letterboxing
    int offsetX = 0, offsetY = 0, xPosition = 0, yPosition = 0;
    if (image_width < contentWidth && image_height < contentHeight)
    {
        offsetX = offsetY = std::min(contentHeight - image_height, widthDifference);
        if (widthDifference > renderLayerPadding)
        {
            xPosition = offsetX / 2;
            yPosition = offsetY / 2;
        }
        else
        {
            yPosition = widthDifference;
            offsetY = widthDifference * 2;
        }
    }

    Color color(0,0,0);
    if (m_buffer)
	{
        int width = (image_width < tvResolutionX) ? image_width : tvResolutionX;
        int height = (image_height < tvResolutionY) ? image_height : tvResolutionY;
        ctx->fillRect(IntRect(xPosition, yPosition, contentWidth - offsetX, contentHeight - offsetY), color, ColorSpaceDeviceRGB);
        ctx->drawImageBuffer(m_buffer.get(),ColorSpaceDeviceRGB,  IntRect(x, y, width, height), IntRect(0, 0, m_buffer->width(), m_buffer->height()));
	}
	else
	{
		pthread_mutex_lock(&m_player_mutex);
		if (m_player)
        {
            m_player->paint(ctx, IntRect(x, y, tvResolutionX, tvResolutionY));
        }
		pthread_mutex_unlock(&m_player_mutex);
	}
}

IntSize VideoLayerWiiU::contentBounds() const
{
   IntSize ret;
   pthread_mutex_lock(&m_player_mutex);
   if (m_player)
      ret = m_player->size();
   pthread_mutex_unlock(&m_player_mutex);
   return ret;
}

bool VideoLayerWiiU::drawsContent() const
{
    return m_player;
}

}
#endif // USE(ACCELERATED_COMPOSITING)
