/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stddef.h>
#include <wtf/Platform.h>
#include <math.h>

#include <PlatformOS.h>
using namespace LibPlatform;

#include "Document.h"
#include "EventHandler.h"
#include "Frame.h"
#include "Node.h"
#include "Page.h"
#include "RenderView.h"
#include "RenderLayer.h"
#include "Element.h"
#include "HTMLElement.h"

#include "WiiUAutoScroller.h"
#include "WiiUScrollConstants.h"

const double WiiUAutoScroller::scrollTimeIncrement = 0.01666666666;

static const float ACC_MAG = 0.95f;
static const float ACC_LEWAY = 0.5f;
float WiiUAutoScroller::slowing = false;

WiiUAutoScroller::WiiUAutoScroller()
	 : m_scrollNode(NULL),
	 m_XVelocity(0.0f),
	 m_YVelocity(0.0f),
	 m_lastTime(0.0f),
	 m_duration(0.0f),
	 m_acceleration(WiiUScrollConstants::default_acceleration()),
	 m_currentStepValue(0.0f),
	 m_startStepValue(0.0f),
	 m_endStepValue(0.0f),
	 m_animationDuration(0.0f),
	 m_animationInterval(0.0f),
	 m_elapsedTime(0.0f),
	 m_isScrolling(false),
	 m_velocityTimer(this, &WiiUAutoScroller::scrollVelocity),
     m_smoothScrollTimer(this, &WiiUAutoScroller::smoothScrollUpdate),
     m_dpadScrollTimer(this, &WiiUAutoScroller::dpadScrollUpdate),
	 m_remainingX(0),
	 m_remainingY(0),
     m_isScrollingForNav(false),
     dragging(false)
{
}

WiiUAutoScroller::~WiiUAutoScroller()
{
	stopScrollWithVelocity();
}

void WiiUAutoScroller::setScrollNode(WebCore::Node *node) 
{ 
	m_scrollNode = node; 
}

void WiiUAutoScroller::startScrollWithVelocity(float x_velocity, float y_velocity)
{
	if (m_acceleration >= 0.0f) {
		m_acceleration = WiiUScrollConstants::default_acceleration();
	}

	m_lastTime = 0.0f;
	m_XVelocity = x_velocity;
	m_YVelocity = y_velocity;

	m_duration = 120;

	//start the timer
	m_startStepValue = m_currentStepValue;
	m_endStepValue = 1.0f;

	if (m_duration == 0) {
		//skip to end of animation
	}else if (m_currentStepValue == m_endStepValue) {
		return;
	}

	m_animationDuration = m_duration * (1 - m_currentStepValue);
	m_animationInterval = calculateInterval(defaultFrameRateHz);
	m_velocityTimer.stop();

	m_isScrolling = true;
	m_isScrollingForNav = false;
	m_velocityTimer.startRepeating(m_animationInterval);
}

void WiiUAutoScroller::stopScrollWithVelocity()
{
	m_XVelocity = m_YVelocity = m_lastTime = m_duration = m_currentStepValue = m_elapsedTime = 0.0f;
	m_acceleration = WiiUScrollConstants::default_acceleration();
	m_startStepValue = 0.0f;
	m_endStepValue = 0.0f;
	m_animationDuration = 0.0f;
	m_animationInterval = 0.0f;
	//stop the timer
	if (isScrolling()) {
		m_velocityTimer.stop();
		m_isScrolling = false;
	}
}

void WiiUAutoScroller::smoothScroll(float vec_x, float vec_y, float vel, float acc, float diffuse, float screen_x, float screen_y)
{
    this->vec_x = vec_x;
    this->vec_y = vec_y;
    this->vel = vel;
    this->acc = acc;
    this->diffuse = diffuse;
    this->screen_x = screen_x;
    this->screen_y = screen_y;
    
    if((diffuse == 1.0f && acc == 0.0f) || !this->m_smoothScrollTimer.isActive()) {
        this->rel_x = 0.0f;
        this->rel_y = 0.0f;

        if(this->m_smoothScrollTimer.isActive()) {
            this->vel = vel * 2.5f;
            this->screen_x = screen_x * 2.5f;
            this->screen_y = screen_y * 2.5f;
        }
    }

    WiiUAutoScroller::slowing = this->diffuse != 1.0f;

	if(!this->m_smoothScrollTimer.isActive()) {
        this->time_start = OS::GetSystemTimeMillisecondsPrecise();
		this->m_smoothScrollTimer.startOneShot(scrollTimeIncrement);
	}
}

void WiiUAutoScroller::smoothScrollUpdate(WebCore::Timer<WiiUAutoScroller>* timer)
{
    double time_elapsed = OS::GetSystemTimeMillisecondsPrecise() - time_start;
    float time_fix = time_elapsed / ((1.0 / 60.0) * 1000.0);

    bool const_vel = (this->diffuse == 1.0f && this->acc == 0.0f);
    bool rel_out_bound = (this->rel_x * this->rel_x > this->screen_x * this->screen_x * 1.5f || this->rel_y * this->rel_y > this->screen_y * this->screen_y * 1.5f);
    float vec_m = this->vec_x * this->vec_x + this->vec_y * this->vec_y;
    
    // If we've wandered farther than half the screen's length from our original
    // position, then start accelerating and accelerations is set, and the stick is maxed out.
    if(!const_vel && rel_out_bound && vec_m > ACC_MAG) {
        float acc_x = this->acc * this->vec_x;
        float acc_y = this->acc * this->vec_y;

        if(this->acc) {
            // if the direction of acceleration changes too much then reset the velocity
            float vel_m = sqrt(this->vel_x * this->vel_x + this->vel_y * this->vel_y);
            float acc_m = sqrt(acc_x * acc_x + acc_y * acc_y);
        
            float vel_x_n = this->vel_x / vel_m;
            float vel_y_n = this->vel_y / vel_m;
            float acc_x_n = acc_x / acc_m;
            float acc_y_n = acc_y / acc_m;
            float cos_theta = vel_x_n * acc_x_n + vel_y_n * acc_y_n;
        
            // If we're drastically changing directions reset relative pos
            if(cos_theta < ACC_LEWAY) {
                this->rel_x = 0.0f;
                this->rel_y = 0.0f;
            }
        }
        
        this->vel_x += acc_x * time_fix;
        this->vel_y += acc_y * time_fix;

        if(this->diffuse == 1.0f && this->acc == 0.0f) {
            this->vel_x = vel * vec_x;
            this->vel_y = vel * vec_y;
        }

    } else /* scroll at constant velocity */ {
        if(rel_out_bound) {
            this->vel_x = 0.0f;
            this->vel_y = 0.0f;
        } else {
            this->vel_x = vel * vec_x;
            this->vel_y = vel * vec_y;
        }
    }

    // If we have any amount of velocity left, scroll the node, update our relative
    // position, and reschedule the update
    scrollNode(this->vel_x, this->vel_y);
    this->rel_x += this->vel_x;
    this->rel_y += this->vel_y;

    // Diffuse velocity (only way to stop this loop)
    // if(!this->acc)
    {
        this->vel_x *= this->diffuse;
        this->vel_y *= this->diffuse;
    }

    if(!WiiUAutoScroller::slowing && (this->vel_x * this->vel_x > 0.5f || this->vel_y * this->vel_y > 0.5f)) {
        m_smoothScrollTimer.startOneShot(scrollTimeIncrement);
    }

    // Kepp track of how long it actually takes to get back from here.
    this->time_start = OS::GetSystemTimeMillisecondsPrecise();
}

void WiiUAutoScroller::dpadScroll(float vec_x, float vec_y, float screen_x, float screen_y)
{
    this->rel_x = 0.0f;
    this->rel_y = 0.0f;
    
	if(!this->m_dpadScrollTimer.isActive()) {
        this->faster = false;
        this->vec_x = vec_x;
        this->vec_y = vec_y;
        this->screen_x = screen_x/2.0f;
        this->screen_y = screen_y/2.0f;
		this->m_dpadScrollTimer.startOneShot(scrollTimeIncrement);
	} else {
        this->faster = true;
    }
}

void WiiUAutoScroller::dpadScrollUpdate(WebCore::Timer<WiiUAutoScroller>* timer)
{
    static const float frames = 10.0f;
    float speed_x = this->screen_x / frames * (this->faster ? 1.5f : 1.0f);
    float speed_y = this->screen_y / frames * (this->faster ? 1.5f : 1.0f);
    float dx = this->vec_x * speed_x;
    float dy = this->vec_y * speed_y;
    this->rel_x += fabs(dx);
    this->rel_y += fabs(dy);
    scrollNode(dx, dy);
    
    if(this->rel_x < this->screen_x && this->rel_y < this->screen_y) {
        m_dpadScrollTimer.startOneShot(scrollTimeIncrement);
    }
}

bool WiiUAutoScroller::shouldScrollWithVelocity(WebCore::Frame *mainFrame, int dx, int dy)
{
	if (!m_scrollNode) {
		return false;
	}

	WebCore::RenderObject* renderer = m_scrollNode->renderer();

	if (!renderer) {
		return false;
	}

	WebCore::RenderLayer* layer = renderer ? renderer->enclosingLayer() : 0;

	if (!layer) {
		return false;
	}

	bool isLineClampNoneByParent = true;

	if (renderer->parent()) {
		isLineClampNoneByParent = renderer->parent()->style()->lineClamp().isNone();
	}

	if (renderer->hasOverflowClip() && isLineClampNoneByParent) {
		int offsetX = layer->scrollXOffset() + dx;
		int offsetY = layer->scrollYOffset() + dy;

		if(offsetX < 0) {
			offsetX = 0;
		}

		if(offsetY < 0) {
			offsetY = 0;
		}

		if(!canScrollLayer(layer, offsetX, offsetY)) {
			setScrollNodeToMainDocument();
			return true;
		}

		return false;
	}

	WebCore::Frame* frame = renderer->frame();

	if (frame == mainFrame) {
		return true;
	}

	if(renderer->view()->frameView() && !canScrollView(renderer->view()->frameView(), dx, dy))
	{
		setScrollNodeToMainDocument();
		return true;
	}

	return false;
}

void WiiUAutoScroller::scrollEnded()
{
	stopScrollWithVelocity();
}

void WiiUAutoScroller::scrollProgressed()
{
	float t = m_currentStepValue * m_duration;
	float a_x = m_XVelocity > 0 ? m_acceleration : -m_acceleration;
	float a_y = m_YVelocity > 0 ? m_acceleration : -m_acceleration;
	float dx = getDelta(m_XVelocity, a_x, m_lastTime, t);
	float dy = getDelta(m_YVelocity, a_y, m_lastTime, t);

	m_XVelocity = m_XVelocity * (1 - m_currentStepValue);
	m_YVelocity = m_YVelocity * (1 - m_currentStepValue);
	m_lastTime = t;
	scrollNode(-1.0f * dx, -1.0f * dy);

	//if the node is no longer actually moving
	if(dx == 0.0f && dy == 0.0f)
	{
		//and the velocity time is active
		if(m_velocityTimer.isActive())
		{
			//Stop the scroll to prevent false positives from isScrolling()
			stopScrollWithVelocity();
		}
	}
}

void WiiUAutoScroller::scrollCanceled()
{
	stopScrollWithVelocity();
}

void WiiUAutoScroller::scrollVelocity(WebCore::Timer<WiiUAutoScroller>* timer)
{
	m_elapsedTime += m_animationInterval;
	m_currentStepValue = m_elapsedTime / m_animationDuration;
	if (m_currentStepValue >= 1.0) {
		m_currentStepValue = 1.0;
	}
	scrollProgressed();
	if (m_currentStepValue == 1.0) {
		stopScrollWithVelocity();
	}
}




float WiiUAutoScroller::calculateInterval(int frameRate)
{
	float timerInterval = 1000000 / frameRate;
	if (timerInterval < 10000) {
		timerInterval = 10000;
	}
	return (timerInterval/1000000);
}

float WiiUAutoScroller::calculateTween(float state)
{
	//We're assuming animation to be ease out
	return 1.0f - pow(1.0f - state, 2);
}

float WiiUAutoScroller::getDelta(float v0, float a, float t1, float t2)
{
	return getPosition(v0, a, t2) - getPosition(v0, a, t1);
}

float WiiUAutoScroller::getPosition(float v0, float a, float t)
{
	float max_t = -v0/a;
	if (t > max_t) {
		t = max_t;
	}

	return t*(v0+0.5f*a*t);
}

void WiiUAutoScroller::scrollNode(int dx, int dy) {

	if(dx == 0 && dy == 0) {
		return;
	}

	if (!m_scrollNode) {
		return;
	}

	WebCore::RenderObject* renderer = m_scrollNode->renderer();
	WebCore::RenderLayer* layer = renderer ? renderer->enclosingLayer() : 0;

	if (!layer) {
		return;
	}

	bool isLineClampNoneByParent = true;
	if (renderer->parent()) {
		isLineClampNoneByParent = renderer->parent()->style()->lineClamp().isNone();
	}
	if (renderer->hasOverflowClip() && isLineClampNoneByParent) {
		WebCore::IntPoint scrollPoint(layer->scrollXOffset() + dx, layer->scrollYOffset() + dy);
		WebCore::IntPoint oldOffset(layer->scrollXOffset(), layer->scrollYOffset());

		//If the targeted layer is big enough to scroll but can't be because of settings
		if(!layer->scrollsOverflow()) {
			//switch scroll node to main frame
			setScrollNodeToMainDocument();
			return;
		}

		WebCore::IntPoint newScrollPosition = scrollPoint.shrunkTo(layer->maximumScrollPosition());
		newScrollPosition = newScrollPosition.expandedTo(layer->minimumScrollPosition());

		if(oldOffset == newScrollPosition) {
            setScrollNodeToMainDocument();
			return;
		}

		layer->scrollToOffset(newScrollPosition.x(), newScrollPosition.y());

		WebCore::Frame* frame = renderer->frame();
		if(frame) {
			WebCore::EventHandler* eventHandler = frame->eventHandler();
			eventHandler->updateAutoscrollRenderer();
		}
	}else if (renderer->view()->frameView()) {
		WebCore::FrameView *fview = renderer->view()->frameView();
 		if(!canScrollView(fview, dx, dy)) {
			if(m_scrollNode->document() && m_scrollNode->document()->page())
			{
				WebCore::Page *page = m_scrollNode->document()->page();
				//if the view being scrolled isn't the main frame
				if(fview != page->mainFrame()->view())
				{
					//continue scrolling main frame
 					if(!m_velocityTimer.isActive())
 						scrollMainFrame(dx,dy);
 					else
 						stopScrollWithVelocity();
 					return;
				}
			}
 		}

		if(m_scrollNode->document())
		{
			WebCore::RenderObject* bodyRenderer = m_scrollNode->document()->body() ? m_scrollNode->document()->body()->renderer() : 0;
			if(bodyRenderer)
			{
				if(bodyRenderer->isBox() )
				{
					//If the body of the page has its overflow CSS properties set, zero out the deltas
					if(toRenderBox(bodyRenderer)->style()->overflowX() == WebCore::OHIDDEN)
						dx = 0;
					if(toRenderBox(bodyRenderer)->style()->overflowY() == WebCore::OHIDDEN)
						dy = 0;
				}
			}
		}

 		renderer->view()->frameView()->scrollBy(WebCore::IntSize(dx,dy));
	}
}


bool WiiUAutoScroller::canScrollLayer(WebCore::RenderLayer *layer, int scrollX, int scrollY)
{
	WebCore::IntPoint scrollMax = layer->maximumScrollPosition();
	WebCore::IntPoint scrollMin = layer->minimumScrollPosition();

	if(!layer->scrollsOverflow())
		return false;

	if(scrollX != 0 && scrollX <= scrollMax.x() && scrollX >= scrollMin.x()) {
		return true;
	}

	if(scrollY != 0 && scrollY <= scrollMax.y() && scrollY >= scrollMin.y()) {
		return true;
	}

	return false;
}

bool WiiUAutoScroller::canScrollView(WebCore::FrameView *view, int dx, int dy)
{
	WebCore::IntPoint offset = view->scrollPosition();
	WebCore::IntPoint scrollMax = view->maximumScrollPosition();
	WebCore::IntPoint scrollMin = view->minimumScrollPosition();

	offset.move(dx, dy);

	if(offset.x() != 0  && offset.x() <= scrollMax.x() && offset.x() >= scrollMin.x()) {
		return true;
	}

	if(offset.y() != 0 && offset.y() <= scrollMax.y() && offset.y() >= scrollMin.y()) {
		return true;
	}

	return false;
}

void WiiUAutoScroller::scrollMainFrame(int dx, int dy)
{
	WebCore::Document *dom = m_scrollNode->document();
	if(!dom) {
		return;
	}

	WebCore::Page *page = dom->page();
	if(!page) {
		return;
	}
	page->mainFrame()->view()->scrollBy(WebCore::IntSize(dx,dy));
}

void WiiUAutoScroller::setScrollNodeToMainDocument()
{
	WebCore::Document *dom = m_scrollNode->document();
	if(!dom) {
		m_scrollNode = NULL;
		return;
	}

	WebCore::Page *page = dom->page();
	if(!page) {
		m_scrollNode = NULL;
		return;
	}
	m_scrollNode = page->mainFrame()->document();
}


