/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include "WiiUFocusAnimator.h"

#include "IntPoint.h"
#include "IntSize.h"
#include "Document.h"
#include "Page.h"
#include "Chrome.h"
#include "Element.h"
#include "Node.h"
#include "ChromeClient.h"
#include "Frame.h"
#include "FrameView.h"
#include "HTMLFrameElement.h"
#include "HTMLNames.h"
#include "FocusController.h"
#include "LengthBox.h"
#include "RenderStyle.h"
#include "NodeRenderStyle.h"

#include "RenderView.h"
#include "RenderLayer.h"
#include "HitTestRequest.h"
#include "HitTestResult.h"
#include "EventHandler.h"

#include <wtf/Forward.h>
#include <wtf/Noncopyable.h>
#include <wtf/RefPtr.h>


WiiUFocusAnimator::WiiUFocusAnimator()
: m_focusChangeTimer(this, &WiiUFocusAnimator::updateChangeFocusRect)
, m_focusFadeTimer(this, &WiiUFocusAnimator::updateFadeFocusRect)
, m_doc (0)
, m_validToNode(false)
, m_validFromNode(false)
, m_client(0)
, m_stepX(0)
, m_stepY(0)
, m_stepW(0)
, m_stepH(0)
, m_numIters(0)
, m_lastEventMouseEvent(false)
, m_focusRectIsInflated(false)
{
	m_isChangingFocus = false;
	m_focusNaved = false;
    m_isFadingFocus = false;
	m_curFocusRect.setLocation(WebCore::FloatPoint(0,0));
	m_curFocusRect.setSize(WebCore::FloatSize(0,0));
    m_curFocusNode = NULL;

	m_endFocusRect.setLocation(WebCore::FloatPoint(0,0));
	m_endFocusRect.setSize(WebCore::FloatSize(0,0));
}


WiiUFocusAnimator::~WiiUFocusAnimator()
{
}

void WiiUFocusAnimator::setDocument(WebCore::Document *doc)
{
	// can also be used to null out the document pointer
	m_doc = doc;
}

void WiiUFocusAnimator::startChangeFocusAnimation(WebCore::Node *to, bool skipToEnd)
{
    m_curFocusNode = to;

	if (to) {
		m_doc = to->document();
	} else {
		
        // If focus navigating into or out of an iFrame during a focus change,
        // then redraw the window
		if(m_isChangingFocus) {
            if(m_client) {
			    m_client->invalidateContentsAndWindow(WebCore::IntRect(m_client->windowRect()), false);
		    }
        }

		m_doc = 0;
		reset();
	}

	if(!m_doc) {
		return;
	}

	WebCore::Page *page = m_doc->page();
	if(!page) {
		return;
	}

	// check if node to is in the current document
	WebCore::LayoutRect toRect = page->focusController()->getNodeRect(to);
	bool validNode = isValidNode(toRect);
	if (!validNode) {
		reset();
		return;
	}
	if(m_isChangingFocus) {
		if(m_focusChangeTimer.isActive()) {
			//m_isChangingFocus = false;
			m_endFocusRect = toRect;
			m_curFocusRect = m_endFocusRect;
			m_focusChangeTimer.stop();
            m_focusChangeTimer.startOneShot(0.008);
		}
		//WebCore::IntRect rect(getCurrentFocusRect());
		if(m_client) {
			m_client->invalidateContentsAndWindow(WebCore::IntRect(m_client->windowRect()), false);
		}
	}

	if(m_focusFadeTimer.isActive()) {
		m_focusFadeTimer.stop();
		m_focusNaved = false;
		m_validToNode = false;
	}

	if(!to && !m_validFromNode) {
		return;
	}

	if(!to) {
		m_validFromNode = false;
		if(!m_lastEventMouseEvent) {
			startFadeFocusAnimation();
		} else {
			m_isChangingFocus = false;
			m_focusNaved = false;
			if(m_client)
				m_client->invalidateContentsAndWindow(WebCore::IntRect(m_client->windowRect()), false);
		}
		return;
	}

	WebCore::Chrome *chrome = page->chrome();
	if(!chrome) {
		return;
	}

	m_client = chrome->client();
	if(!m_client) {
		return;
	}

	m_validToNode = true;

    WebCore::FloatRect visibleRect = page->mainFrame()->view()->visibleContentRect(/* includeScrollbars */ false);

	if(!m_endFocusRect.isEmpty() && visibleRect.intersects(m_endFocusRect)) {
        // if endFocusRect exists and is visible on the current page, start animating from there
		m_startFocusRect = m_endFocusRect;
	} else if (skipToEnd ||
            to->hasTagName(WebCore::HTMLNames::inputTag) ||
            to->hasTagName(WebCore::HTMLNames::selectTag) ||
            to->hasTagName(WebCore::HTMLNames::textareaTag)) {
		m_startFocusRect = toRect;
  	} else {
        // otherwise, animate in on toRect
		m_startFocusRect = toRect;
		m_startFocusRect.inflateX(toRect.width()/2.0);
        m_startFocusRect.inflateY(toRect.height());
        m_focusRectIsInflated = true;
	}

    if (to->parent()->hasTagName(WebCore::HTMLNames::divTag)) {

        bool isScrollable = false;
        WebCore::IntRect divRect = to->parent()->getRect();
	    WebCore::RenderObject* renderer = to->parent()->renderer();
        WebCore::RenderLayer* layer = toRenderBox(renderer)->layer();
        if (layer) {
            isScrollable = layer->allowsScrolling();
        }

        if (!divRect.contains(toRect) && isScrollable) {
            // for scrollable div objects, if the new target is not yet visible we set the target to 
            // the current location. This should correct itself after the scroll
            // and will prevent the focus ring from being drawn outside the div object
            toRect = WebCore::IntRect(m_endFocusRect);
            return;
        }
    }

	WebCore::RenderStyle *style = to->renderStyle();
	if(style) {
		m_endFocusRect.setX((int)toRect.x() + style->borderLeftWidth());
		m_endFocusRect.setY((int)toRect.y() + style->borderTopWidth());
		m_endFocusRect.setWidth((int)toRect.width() - style->borderLeftWidth() - style->borderRightWidth());
		m_endFocusRect.setHeight((int)toRect.height()- style->borderTopWidth() - style->borderBottomWidth());

        if (skipToEnd ||
                to->hasTagName(WebCore::HTMLNames::inputTag) ||
                to->hasTagName(WebCore::HTMLNames::selectTag) ||
                to->hasTagName(WebCore::HTMLNames::textareaTag)) {
            m_startFocusRect = m_endFocusRect;
        }
	} else {
		m_endFocusRect = toRect;
	}

    if (to->hasTagName(WebCore::HTMLNames::divTag)) {
		m_startFocusRect = toRect;
		m_endFocusRect = m_startFocusRect;
    }

	m_focusNaved = true;

    if(!m_isChangingFocus) {
    	m_curFocusRect = m_startFocusRect;
    }

	m_stepX = (float)(m_endFocusRect.x() - m_curFocusRect.x()) / 12.0;
	m_stepY = (float)(m_endFocusRect.y() - m_curFocusRect.y()) / 12.0;
	m_stepW = (float)(m_endFocusRect.width() - m_curFocusRect.width()) / 12.0;
	m_stepH = (float)(m_endFocusRect.height() - m_curFocusRect.height()) / 12.0;

	m_isChangingFocus = true;
	if(!m_focusChangeTimer.isActive()) {
		m_focusChangeTimer.startOneShot(0.008);
	}
}

void WiiUFocusAnimator::updateChangeFocusRect(WebCore::Timer<WiiUFocusAnimator> *timer)
{
	float incX = 0.0;
	float incY = 0.0;
	float incW = 0.0;
	float incH = 0.0;

	if(!m_doc) {
		return;
	}

	if (!m_validToNode) {
		return;
	}

	// if the node for m_endFocusRect is not in the current page ... stop
	bool validNode = isValidNode(m_endFocusRect);
	if (!validNode) {
		reset();
		return;
	}


	WebCore::Page *page = m_doc->page();
	if(!page) {
		return;
	}

	m_damageRect.setX(m_curFocusRect.x());
	m_damageRect.setY(m_curFocusRect.y());
	m_damageRect.setWidth(m_curFocusRect.width());
	m_damageRect.setHeight(m_curFocusRect.height());


	if(fabs(m_endFocusRect.x() - m_curFocusRect.x()) <= fabs(m_stepX)) {
		m_curFocusRect.setX(m_endFocusRect.x());
	} else {
		incX = m_stepX;
	}

	if(fabs(m_endFocusRect.y() - m_curFocusRect.y()) <= fabs(m_stepY)) {
			m_curFocusRect.setY(m_endFocusRect.y());
	} else {
		incY = m_stepY;
	}

	if(fabs(m_endFocusRect.width() - m_curFocusRect.width()) <= fabs(m_stepW)) {
			m_curFocusRect.setWidth(m_endFocusRect.width());
	} else {
		incW = m_stepW;
	}

	if(fabs(m_endFocusRect.height() - m_curFocusRect.height()) <= fabs(m_stepH)) {
			m_curFocusRect.setHeight(m_endFocusRect.height());
	} else {
		incH = m_stepH;
	}

	int offX = 0;
	int offY = 0;

	WebCore::FrameView *view = page->mainFrame()->view();
	if(view) {
		offX =  view->scrollX();
		offY =  view->scrollY();
	}

    int damageSize = 15;

	if (m_curFocusRect == m_endFocusRect) {
		m_isChangingFocus = false;
		m_curFocusRect = m_endFocusRect;
		m_validFromNode = true;
        damageSize = 200; // redraw a larger area to handle scrolling inside divs
	} else {
		m_curFocusRect.move(incX, incY);
		m_curFocusRect.expand(incW, incH);
		m_focusChangeTimer.startOneShot(0.008);
	}

	m_damageRect.unite(WebCore::IntRect(m_curFocusRect));
	m_damageRect.inflate(damageSize);
	m_damageRect.move(-offX, -offY);
	m_client->invalidateContentsAndWindow(m_damageRect, false);
}

void WiiUFocusAnimator::startFadeFocusAnimation()
{
	if(!m_focusNaved || m_isChangingFocus) {
		return;
	}

	if(!m_validToNode) {
		return;
	}

	if(!m_doc) {
		return;
	}

    if (m_isFadingFocus) {
        return;
    }

    m_numIters = 0;
	m_isChangingFocus = false;
	m_focusNaved = true;
    m_isFadingFocus = true;

	if(!m_focusFadeTimer.isActive()) {
		m_focusFadeTimer.startOneShot(0.008);
	}
}

void WiiUFocusAnimator::updateFadeFocusRect(WebCore::Timer<WiiUFocusAnimator> *timer)
{
	float offX = 0;
	float offY = 0;

	if(!m_doc) {
		return;
	}

	WebCore::FrameView *view = m_doc->view();
	if(!view) {
		return;
	}

	WebCore::Page *page = m_doc->page();
	if(!page) {
		return;
	}

	offX =  view->scrollX();
	offY =  view->scrollY();

	m_damageRect = WebCore::IntRect(m_curFocusRect);
	m_curFocusRect.inflate(1);
	m_damageRect.unite(WebCore::IntRect(m_curFocusRect));
	m_damageRect.move(-offX, -offY);
	m_damageRect.inflate(50);
    m_client->invalidateContentsAndWindow(m_damageRect, false);

	if(m_numIters >= 10) {
		m_focusNaved = false;
		m_validToNode = false;
		m_validFromNode = false;
		m_curFocusRect.setX(0.0);
		m_curFocusRect.setY(0.0);
		m_curFocusRect.setWidth(0.0);
		m_curFocusRect.setHeight(0.0);
        m_isFadingFocus = false;
	} else {
		m_numIters++;
		m_focusFadeTimer.startOneShot(0.0008);
	}
}

WebCore::FloatRect WiiUFocusAnimator::getCurrentFocusRect()
{
	if(!m_doc) {
		return WebCore::FloatRect();
	}

	WebCore::Page *page = m_doc->page();
	if(!page) {
		return WebCore::FloatRect();
	}

	WebCore::FrameView *view = page->mainFrame()->view();
	if(!view) {
		return WebCore::FloatRect();
	}

	int offX =  view->scrollX();
	int offY =  view->scrollY();

    if (!m_isChangingFocus && !m_isFadingFocus) {
        // get the rect from m_curFocusNode in case the node has moved
    	WebCore::LayoutRect rect = page->focusController()->getNodeRect(m_curFocusNode);

        if (m_curFocusNode->parent()->hasTagName(WebCore::HTMLNames::divTag)) {

            WebCore::IntRect divRect = m_curFocusNode->parent()->getRect();
            if (divRect.intersects(rect)) {
                // update m_endFocusRect too, unless it is inside a div and not currently visible
                if(!m_endFocusRect.isEmpty()) {
	                m_endFocusRect = rect;
                }
		        rect.move(-offX, -offY);
                return rect;
            }
        } else {
            // update m_endFocusRect too, if necessary
            if(!m_endFocusRect.isEmpty()) {
	            m_endFocusRect = rect;
            }

		    rect.move(-offX, -offY);
            return rect;
        }
    }

    WebCore::IntRect rect(m_curFocusRect);
	rect.move(-offX, -offY);
	return rect;
}

void WiiUFocusAnimator::reset() {
	m_validFromNode = false;
	m_validToNode = false;

	if(m_focusFadeTimer.isActive()) {
		m_focusFadeTimer.stop();
	}
	m_focusNaved = false;
    m_isFadingFocus = false;

	if(m_focusChangeTimer.isActive()) {
		m_focusChangeTimer.stop();
	}
	m_isChangingFocus = false;

	m_curFocusRect.setLocation(WebCore::FloatPoint(0,0));
	m_curFocusRect.setSize(WebCore::FloatSize(0,0));
    m_curFocusNode = NULL;

	m_endFocusRect.setLocation(WebCore::FloatPoint(0,0));
	m_endFocusRect.setSize(WebCore::FloatSize(0,0));
}

void WiiUFocusAnimator::FrameClosing() {
    if (m_doc && m_doc->page()) {
        if (m_doc->page()->focusController()->focusedOrMainFrame() != m_doc->page()->mainFrame()) {
            // if we are inside an iframe, reset the focus animator
            reset();
        }
    }
	m_doc = 0;
}

bool WiiUFocusAnimator::isValidNode(WebCore::FloatRect rect) {
	if (rect.isEmpty()) {
		return false;
	}

	WebCore::IntRect integerRect(rect);
	WebCore::Node* node = getNodeFromPoint((int)rect.x(), (int)rect.y());

	if (node) {
		return true;
	}

	return false;
}

WebCore::Node* WiiUFocusAnimator::getNodeFromPoint(int x, int y) {

	WebCore::Node* node = 0;

	if(!m_doc) {
		return 0;
	}
	WebCore::Page *page = m_doc->page();
	if(!page) {
		return 0;
	}
	WebCore::Frame* frame = page->mainFrame();
	if (!frame) {
		return 0;
	}

	while(frame) {
		WebCore::Document* doc = frame->document();
		WebCore::FrameView* view = frame->view();
		WebCore::IntPoint documentPoint = view ? view->windowToContents(WebCore::IntPoint(x, y)) : WebCore::IntPoint(x, y);
		WebCore::RenderView* renderView = doc->renderView();
		WebCore::HitTestRequest request(WebCore::HitTestRequest::ReadOnly | WebCore::HitTestRequest::Active);
		WebCore::HitTestResult result(documentPoint);

		renderView->layer()->hitTest(request, result);
		node = result.innerNode();

		while (node && !node->isElementNode()) {
			node = node->parentNode();
		}
		if (node && node->isShadowRoot() && node->shadowHost()) {
			node = node->shadowHost();
		}
		frame = WebCore::EventHandler::subframeForTargetNode(node);
	}

	if (!node) {
		return 0;
	}

	return node;
}

void WiiUFocusAnimator::setLastEventWasMouseEvent(bool lastEventMouseEvent) { 
    m_lastEventMouseEvent = lastEventMouseEvent;
    if (m_lastEventMouseEvent) {
		if(m_client) {
			m_client->invalidateContentsAndWindow(WebCore::IntRect(m_client->windowRect()), false);
		}
	}
}

void WiiUFocusAnimator::redrawWindow() { 
	if(m_client) {
		m_client->invalidateContentsAndWindow(WebCore::IntRect(m_client->windowRect()), false);
	}
}