/*
 * Copyright (c) 2012 NINTENDO.
 *
 * Redistribution and use in source and binary forms, with or without
 * modification, are permitted provided that the following conditions
 * are met:
 *
 * 1. Redistributions of source code must retain the above copyright
 *    notice, this list of conditions, and the following disclaimer.
 * 2. Redistributions in binary form must reproduce the above copyright
 *    notice, this list of conditions, and the following disclaimer in the
 *    documentation and/or other materials provided with the distribution.
 *
 * THIS SOFTWARE IS PROVIDED BY NINTENDO AND ITS CONTRIBUTORS "AS IS" AND ANY
 * EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
 * WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE
 * ARE DISCLAIMED.  IN NO EVENT SHALL NINTENDO OR CONTRIBUTORS BE LIABLE FOR ANY
 * DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES
 * (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES;
 * LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED
 * AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
 * (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF
 * THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
 */

#include <stddef.h>
#include <wtf/Platform.h>

#include <PlatformOS.h>
using namespace LibPlatform;

#include "WiiUScrollPoint.h"
#include "WiiUScrollConstants.h"


WiiUScrollPoint::WiiUScrollPoint()
	: m_firstTouchTime(0),
	  m_secondLastTouchTime(0),
	  m_lastTouchTime(0),
	  velocity_calculator(9)
{
}

WiiUScrollPoint::~WiiUScrollPoint() {}

void WiiUScrollPoint::Reset() {
	m_firstTouchTime = m_secondLastTouchTime = m_lastTouchTime = 0.0;
	ResetVelocity();
}

void WiiUScrollPoint::ResetVelocity() {
	velocity_calculator.ClearHistory();
	same_direction_count = WebCore::IntSize();
}

WebCore::IntSize WiiUScrollPoint::ScrollDelta()  {
	return m_lastTouchPosition - m_secondLastTouchPosition;
}

void WiiUScrollPoint::UpdateValues(WebCore::MouseEventType type, const WebCore::PlatformMouseEvent& ev) {
	const long long event_timestamp_microseconds = ev.timestamp();
	if (type == WebCore::MouseEventMoved) {
		velocity_calculator.PointSeen(ev.x(), ev.y(), event_timestamp_microseconds);
		WebCore::IntSize sd(ScrollVelocityDirection(velocity_calculator.XVelocity()),
							ScrollVelocityDirection(velocity_calculator.YVelocity()));
		same_direction_count = same_direction_count + sd;
		m_secondLastTouchPosition = m_lastTouchPosition;
		m_secondLastTouchTime = m_lastTouchTime;
	}

	m_lastTouchTime = ev.timestamp();
	m_lastTouchPosition = WebCore::IntPoint(ev.x(), ev.y());

	if (type == WebCore::MouseEventPressed) {
		ResetVelocity();
		m_firstTouchTime = m_lastTouchTime;
		m_firstTouchPosition = m_lastTouchPosition;
		m_secondLastTouchTime = m_lastTouchTime;
		m_secondLastTouchPosition = m_lastTouchPosition;
		velocity_calculator.PointSeen(ev.x(), ev.y(), event_timestamp_microseconds);
	}
}

void WiiUScrollPoint::UpdateForScroll() {
	m_secondLastTouchPosition = m_lastTouchPosition;
	m_secondLastTouchTime = m_lastTouchTime;
	same_direction_count = WebCore::IntSize();
}

int WiiUScrollPoint::ScrollVelocityDirection(float v) {
	if (v < -WiiUScrollConstants::min_scroll_velocity()) {
		return -1;
	}else if (v > WiiUScrollConstants::min_scroll_velocity()) {
		return 1;
	}else {
		return 0;
	}
}

bool WiiUScrollPoint::IsInFlickWindow() {
	return IsOverMinFlickSpeed();
}

bool WiiUScrollPoint::DidScroll(const WiiUMouseEvent& ev, int dist)  {
	WebCore::IntSize d = m_lastTouchPosition - m_secondLastTouchPosition;
	return (abs(d.width()) > dist || abs(d.height()) > dist);
}

bool WiiUScrollPoint::IsConsistentScrollingActionUnderway()  {
	int me = WiiUScrollConstants::min_scroll_successive_velocity_events();
	if (abs(same_direction_count.width()) >= me ||
		abs(same_direction_count.height()) >= me) {
		return true;
	}
	return false;
}

bool WiiUScrollPoint::IsInHorizontalRailWindow()  {
	WebCore::IntSize d = m_lastTouchPosition - m_secondLastTouchPosition;
	return abs(d.width()) > WiiUScrollConstants::rail_start_proportion() * abs(d.height());
}

bool WiiUScrollPoint::IsInVerticalRailWindow()  {
	WebCore::IntSize d = m_lastTouchPosition - m_secondLastTouchPosition;
	return abs(d.height()) > WiiUScrollConstants::rail_start_proportion() * abs(d.width());
}

bool WiiUScrollPoint::BreaksHorizontalRail() {
	float vx = XVelocity();
	float vy = YVelocity();
	return fabs(vy) > WiiUScrollConstants::rail_break_proportion() * fabs(vx);
}

bool WiiUScrollPoint::BreaksVerticalRail() {
	float vx = XVelocity();
	float vy = YVelocity();
	return fabs(vx) > WiiUScrollConstants::rail_break_proportion() * fabs(vy);
}

bool WiiUScrollPoint::IsOverMinFlickSpeed() {
	return velocity_calculator.VelocitySquared() > WiiUScrollConstants::min_flick_speed_squared();
}

VelocityCalculator::VelocityCalculator(int bufferSize)
	:buffer(new Point[bufferSize]),
	 index(0),
	 num_valid_entries(0),
	 buffer_size(bufferSize),
	 x_velocity(0),
	 y_velocity(0),
	 velocities_stale(false)
{
}

VelocityCalculator::~VelocityCalculator()
{
	delete buffer;
}

float VelocityCalculator::XVelocity() {
	if (velocities_stale) {
		UpdateVelocity();
	}
	return x_velocity;
}

float VelocityCalculator::YVelocity() {
	if (velocities_stale) {
		UpdateVelocity();
	}
	return y_velocity;
}

void VelocityCalculator::PointSeen(int x, int y, long long time) {
	buffer[index].x = x;
	buffer[index].y = y;
	buffer[index].time = time;

	index = (index + 1) % buffer_size;
	if (num_valid_entries < buffer_size) {
		++num_valid_entries;
	}

	velocities_stale = true;
}

float VelocityCalculator::VelocitySquared() {
	if (velocities_stale) {
		UpdateVelocity();
	}
	return x_velocity * x_velocity + y_velocity * y_velocity;
}

void VelocityCalculator::UpdateVelocity() {
	//We don't have enough data to make a good estimate of the velocity
	if (num_valid_entries < 2) {
		return;
	}

	//Where A_i = A[i] - mean(A)
	//x velocity = sum_i(x_i * t_i) / sum_i(t_i * t_i)
	//this is an Ordinary Least Square Regression
	float mean_x = 0;
	float mean_y = 0;
	float mean_time = 0;

	for (size_t i = 0; i < num_valid_entries; i++) {
		mean_x += buffer[i].x;
		mean_y += buffer[i].y;
		mean_time += buffer[i].time;
	}

	//Minimize number of divides
	const float num_valid_entries_i = 1.0f / num_valid_entries;

	mean_x *= num_valid_entries_i;
	mean_y *= num_valid_entries_i;

	mean_time /= num_valid_entries;

	float xt = 0; //sum_i(x_i * t_i)
	float yt = 0; //sum_i(y_i * t_i)
	float tt = 0;  //sum_i(t_i * t_i)

	float t_i;

	for (size_t i = 0; i < num_valid_entries; i++) {
		t_i = (buffer[i].time - mean_time);
		xt += (buffer[i].x - mean_x) * t_i;
		yt += (buffer[i].y - mean_y) * t_i;
		tt += t_i * t_i;
	}

	if (tt > 0) {
		//Convert time from milliseconds to seconds
		x_velocity = xt / (tt / 1000000.0f);
		y_velocity = yt / (tt / 1000000.0f);
	}else {
		x_velocity = 0;
		y_velocity = 0;
	}
	velocities_stale = false;
}

void VelocityCalculator::ClearHistory() {
	index = 0;
	num_valid_entries = 0;
	x_velocity = 0;
	y_velocity = 0;
	velocities_stale = false;
}



